"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import os
import sys
from collections.abc import Mapping, Sequence

import ruyaml as yaml

from .dict import attrdict

try:
    from .msgpack import Proxy
except ImportError:
    Proxy = None
try:
    from moat.lib.codec.proxy import DProxy
except ImportError:
    DProxy = None

from .path import Path

__all__ = [
    "yload",
    "yprint",
    "yformat",
    "yaml_repr",
    "yaml_parse",
    "add_repr",
    "load_ansible_repr",
]

SafeRepresenter = yaml.representer.SafeRepresenter
SafeConstructor = yaml.constructor.SafeConstructor
Emitter = yaml.emitter.Emitter


SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)


def load_ansible_repr():
    "Call me if you're using `moat.util` in conjunction with Ansible."
    from ansible.parsing.yaml.objects import AnsibleUnicode
    from ansible.utils.unsafe_proxy import AnsibleUnsafeText
    from ansible.vars.hostvars import HostVars, HostVarsVars

    SafeRepresenter.add_representer(HostVars, SafeRepresenter.represent_dict)
    SafeRepresenter.add_representer(HostVarsVars, SafeRepresenter.represent_dict)
    SafeRepresenter.add_representer(AnsibleUnsafeText, SafeRepresenter.represent_str)
    SafeRepresenter.add_representer(AnsibleUnicode, SafeRepresenter.represent_str)


def str_presenter(dumper, data):
    """
    Always show multi-line strings with |-style quoting
    """
    if "\n" in data:  # multiline string?
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    else:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def yaml_repr(name: str, use_repr: bool = False):
    """
    A class decorator that allows representing an object in YAML
    """

    def register(cls):
        def str_me(dumper, data):
            return dumper.represent_scalar("!" + name, repr(data) if use_repr else str(data))

        SafeRepresenter.add_representer(cls, str_me)
        return cls

    return register


def yaml_parse(name: str, use_repr: bool = False):
    """
    A decorator that allows parsing a YAML representation,
    i.e. the opposite of `yaml_repr`.
    """

    def register(cls):
        SafeConstructor.add_constructor(f"!{name}", cls)
        return cls

    use_repr  # noqa: B018
    return register


SafeRepresenter.add_representer(str, str_presenter)


def _path_repr(dumper, data):
    return dumper.represent_scalar("!P", str(data))
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))


def _proxy_repr(dumper, data):
    return dumper.represent_scalar("!R", data.name)
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))

def _dproxy_repr(dumper, data):
    return dumper.represent_scalar("!DP", repr([data.name,data.a,data.k]))
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))


def read_env(loader, node):
    value = loader.construct_scalar(node)
    return os.environ[value]

def _el_repr(dumper, data):
    return dumper.represent_scalar("!R", "...")


SafeRepresenter.add_representer(type(Ellipsis), _el_repr)
SafeRepresenter.add_representer(Path, _path_repr)
SafeConstructor.add_constructor("!P", Path._make)

SafeConstructor.add_constructor("!env", read_env)

if Proxy is not None:
    SafeRepresenter.add_representer(Proxy, _proxy_repr)
if DProxy is not None:
    SafeRepresenter.add_representer(DProxy, _dproxy_repr)


def _name2obj(constructor, node):
    from moat.lib.codec.proxy import name2obj
    return name2obj(node.value)


SafeConstructor.add_constructor("!R", _name2obj)


def _bin_from_ascii(loader, node):
    value = loader.construct_scalar(node)
    return value.encode("ascii")


def _bin_from_hex(loader, node):
    value = loader.construct_scalar(node)
    return bytearray.fromhex(value.replace(":", ""))


def _bin_to_ascii(dumper, data):
    try:
        data = data.decode("utf-8")
    except UnicodeError:
        if len(data) < 33:
            return dumper.represent_scalar("!hex", data.hex(":"))
        else:
            return dumper.represent_binary(data)
    else:
        return dumper.represent_scalar("!bin", data)


SafeRepresenter.add_representer(bytes, _bin_to_ascii)
SafeRepresenter.add_representer(bytearray, _bin_to_ascii)
SafeRepresenter.add_representer(memoryview, _bin_to_ascii)

SafeConstructor.add_constructor("!bin", _bin_from_ascii)
SafeConstructor.add_constructor("!hex", _bin_from_hex)


_expect_node = Emitter.expect_node


def expect_node(self, *a, **kw):
    """
    YAML stream mangler.

    TODO rationale?
    """
    _expect_node(self, *a, **kw)
    self.root_context = False


Emitter.expect_node = expect_node


def yload(stream, multi=False, attr=False):
    """
    Load one or more YAML objects from a file.
    """
    y = yaml.YAML(typ="safe")
    if attr:

        class AttrConstructor(SafeConstructor):  # pylint: disable=missing-class-docstring
            def __init__(self, *a, **k):
                super().__init__(*a, **k)
                self.yaml_base_dict_type = attrdict if attr is True else attr

        y.Constructor = AttrConstructor
    if multi:
        return y.load_all(stream)
    else:
        return y.load(stream)


def yprint(data, stream=sys.stdout, compact=False):
    """
    Write a YAML record.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    if isinstance(data, (int, float)):
        print(data, file=stream)
    elif isinstance(data, (str, bytes)):
        print(repr(data), file=stream)
    #   elif isinstance(data, bytes):
    #       os.write(sys.stdout.fileno(), data)
    else:
        y = yaml.YAML(typ="safe")
        y.default_flow_style = compact
        y.dump(data, stream=stream)


def yformat(data, compact=None):
    """
    Return ``data`` as a multi-line YAML string.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    from io import StringIO  # pylint: disable=import-outside-toplevel

    s = StringIO()
    yprint(data, compact=compact, stream=s)
    return s.getvalue()


def add_repr(typ, r=None):
    """
    Add a way to add representations for subtypes.

    This is useful for subclassed dict/int/str/… objects.
    """
    # pylint: disable=redefined-builtin
    if r is None:
        r = typ
    if issubclass(r, str):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_str)
    elif issubclass(r, float):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_float)
    elif issubclass(r, bool):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_bool)
    elif issubclass(r, int):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_int)
    elif issubclass(r, Mapping):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_dict)
    elif issubclass(r, Sequence):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_list)
    else:
        raise TypeError(f"Don't know what to do with {typ}")



"""
trio/anyio no longer have queues, but sometimes a memory object stream is
unwieldy. Thus this re-implemens a simple queue using
`anyio.create_memory_object_stream`.
"""

from __future__ import annotations

import anyio
from anyio import create_memory_object_stream as _cmos

from outcome import Error, Value

from .dict import attrdict

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from collections.abc import Awaitable

import logging  # isort:skip

logger = logging.getLogger(__name__)

__all__ = [
    "Queue",
    "QueueFull",
    "QueueEmpty",
    "create_queue",
    "DelayedWrite",
    "DelayedRead",
]

QueueFull = anyio.WouldBlock
QueueEmpty = anyio.WouldBlock


class Queue:
    """
    Queues have been replaced in trio/anyio by memory object streams, but
    those are more complicated to use.

    This Queue class simply re-implements queues on top of memory object streams.
    """

    def __init__(self, length=0):
        self._s, self._r = _cmos(length)

    @property
    def _moat(self):
        try:
            return self.__moat
        except AttributeError:
            self.__moat = d = attrdict()  # pylint: disable=attribute-defined-outside-init
            return d

    async def put(self, x):
        """Send a value, blocking"""
        try:
            await self._s.send(Value(x))
        except anyio.ClosedResourceError:
            raise EOFError from None

    def put_nowait(self, x):
        """Send a value, nonblocking"""
        try:
            self._s.send_nowait(Value(x))
        except anyio.ClosedResourceError:
            raise EOFError from None

    async def put_error(self, x):
        """Send an error value, blocking"""
        try:
            await self._s.send(Error(x))
        except anyio.ClosedResourceError:
            raise EOFError from None

    def put_nowait_error(self, x):
        """Send an error, nonblocking"""
        try:
            self._s.send_nowait(Error(x))
        except anyio.ClosedResourceError:
            raise EOFError from None

    async def get(self):
        """Get the next value, blocking.
        May raise an exception if one was sent."""
        try:
            res = await self._r.receive()
        except anyio.EndOfStream:
            raise EOFError from None
        return res.unwrap()

    def get_nowait(self):
        """Get the next value, nonblocking.
        May raise an exception if one was sent."""
        try:
            res = self._r.receive_nowait()
        except anyio.EndOfStream:
            raise EOFError from None
        return res.unwrap()

    def qsize(self):
        """Return the number of elements in the queue"""
        return self._s.statistics().current_buffer_used

    def empty(self):
        """Check whether the queue is empty"""
        return self._s.statistics().current_buffer_used == 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        res = await self._r.__anext__()  # pylint: disable=E1101
        return res.unwrap()

    def close_sender(self) -> Awaitable:
        """No more messages will be received"""
        self._s.close()

    close_writer = close_sender

    def close_receiver(self) -> Awaitable:
        """No more messages may be sent"""
        self._r.close()

    close_reader = close_receiver


class Lockstep(Queue):
    "A queue with sender/receiver rendez-vous"

    def __init__(self):
        super().__init__(0)


def create_queue(length=0):
    """Create a queue. Compatibility method.

    Deprecated; instantiate `Queue` directly."""
    return Queue(length)


class DelayedWrite:
    """
    A module that limits the number of outstanding outgoing messages by
    receiving flow-control messages from a `DelayedRead` instance on the
    other side.
    """

    _delay = None
    _send_lock = None
    _info = None
    _seq = 0

    def __init__(self, length, info=None):
        self.len = length
        self._n_ack = 0
        self._n_sent = 0
        self._send_lock = anyio.Lock()
        if info is None:
            DelayedWrite._seq += 1
            info = f"DlyW.{DelayedWrite._seq}"
        self._info = info

    async def next_seq(self):
        """
        Returns the next seq num for sending.

        May delay until an ack is received.
        """
        async with self._send_lock:
            self._n_sent += 1
            res = self._n_sent
            if self._delay is None and self._n_sent - self._n_ack >= self.len:
                logger.debug("%s: wait: %d/%d", self._info, self._n_sent, self._n_ack)
                self._delay = anyio.Event()
            if self._delay is not None:
                await self._delay.wait()
            return res

    async def recv_ack(self, msg_nr):
        """
        Signal that this ack msg has been received.
        """
        self._n_ack = max(self._n_ack, msg_nr)
        if self._delay is not None and self._n_sent - self._n_ack < self.len:
            logger.debug("%s: go: %d/%d", self._info, self._n_sent, self._n_ack)
            self._delay.set()
            self._delay = None


class DelayedRead(Queue):
    """
    A queue that limits the number of outstanding incoming messages by
    flow-controlling a `DelayedWrite` instance on the other side.

    You need to override (or pass in)

    * get_seq(msg) -- extract the msgnum from a message
    * async send_ack(seq) -- send an ack for this message
    """

    def __init__(self, length, *, get_seq=None, send_ack=None):
        if length < 4:
            raise RuntimeError("Length <4 doesn't make sense")
        super().__init__(length)
        self._n_last = 0
        self._n_ack = 0
        self._len = length // 3
        if get_seq is not None:
            self.get_seq = get_seq
        if send_ack is not None:
            self.send_ack = send_ack

    @staticmethod
    def get_seq(msg) -> int:  # pylint: disable=method-hidden
        """msgnum extractor. Static method. Override me!"""
        raise NotImplementedError("Override get_seq")

    async def send_ack(self, seq: int):  # pylint: disable=method-hidden
        """Ack sender for a specific seqnum. Override me!"""
        raise NotImplementedError("Override send_flow")

    async def _did_read(self, res):
        self._n_last = max(self._n_last, self.get_seq(res))
        if self._n_last - self._n_ack > self._len:
            self._n_ack = self._n_last
            await self.send_ack(self._n_last)

    async def __anext__(self):
        res = await super().__anext__()
        await self._did_read(res)
        return res

    async def get(self):
        """Receive the next message, send an Ack to the other side."""
        res = await super().get()
        await self._did_read(res)
        return res



"""
This module affords a helper class to make "async with OBJECT" map
seamlessly to an async context management method.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from attrs import define
import anyio

from typing import TYPE_CHECKING, overload, TypeVar

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable
    from types import TracebackType


__all__ = ["CtxObj", "timed_ctx"]

T_Ctx = TypeVar("T_Ctx")


class CtxObj(ABC):
    """
    Teach a class instance to act as an async context manager, by
    forwarding ``_aenter__````_aexit__`` to a `_ctx` method
    (which must be an `AsyncContextManager`).

    Usage::
        class Foo(CtxObj):
            @asynccontextmanager
            async def _ctx(self):
                yield self # or whatever

        async with Foo() as whatever:
            pass
    """

    __ctx: AbstractAsyncContextManager | None = None

    @abstractmethod
    def _ctx(self) -> AsyncIterator[T_Ctx]:
        ...

    async def __aenter__(self) -> T_Ctx:
        if self.__ctx is not None:
            raise RuntimeError("Nested contexts")
        ctx = self._ctx()
        if not hasattr(ctx,"__aenter__"):
            # DEPRECATED
            # legacy code for `_ctx` without @asynccm
            ctx = asynccontextmanager(self._ctx)()
        self.__ctx = ctx
        return await ctx.__aenter__()

    def __aexit__(
        self,
        *tb: *tuple[type[BaseException] | None, BaseException | None, TracebackType | None],
    ) -> Awaitable[bool | None]:
        try:
            assert self.__ctx is not None
            return self.__ctx.__aexit__(*tb)
        finally:
            self.__ctx = None


@define
class timed_ctx(CtxObj):
    """
    A wrapper for an async context manager that times out if entering it
    takes too long.

    Everything else is unaffected.
    """

    timeout: int | float
    mgr: AbstractAsyncContextManager

    async def _timer(self, *, task_status):
        with anyio.CancelScope() as sc:
            task_status.started(sc)
            await anyio.sleep(self.timeout)
            raise TimeoutError(self.timeout)

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            sc = await tg.start(self._timer)
            async with self.mgr as mgr:
                sc.cancel()
                yield mgr



"""
Support for main program, argv, etc.
"""

from __future__ import annotations

import importlib
import logging
import logging.config
import os
import sys
from collections import defaultdict
from collections.abc import Mapping
from contextlib import suppress
from contextvars import ContextVar
from functools import partial, wraps
from pathlib import Path as FSPath

import asyncclick as click

from .dict import attrdict, to_attrdict
from .exc import ungroup
from . import NotGiven
from .merge import merge
from .config import CFG, ensure_cfg

try:
    from .msgpack import Proxy
except ImportError:
    Proxy = None
from .path import P
from .yaml import yload

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Awaitable  # noqa:UP035
    from asyncclick import command

logger = logging.getLogger("_loader")

__all__ = [
    "main_",
    "read_cfg",
    "load_cfg",
    "wrap_main",
    "Loader",
    "load_subgroup",
    "list_ext",
    "load_ext",
    "attr_args",
    "process_args",
    "option_ng",
]

this_load = ContextVar("this_load", default=None)

NoneType = type(None)


def _no_config(*a, **k):  # noqa:ARG001
    import warnings

    warnings.warn("Call to logging config ignored", stacklevel=2)


def attr_args(
    proc=None,
    with_combined="s",
    with_path=True,
    with_eval=True,
    with_var=True,
    with_proxy=False,
    par_name="Parameter",
):
    """
    Add an option for setting possibly-hierarchical values to `click.command`.

    New behavior (default): use ``-s``/``--set`` with prefix-tagged values:
    * ~str
    * =value (``=-``/``=t``/``=f``/``=n`` for ``undefined``/`True`/`False`/`None`)
    * .path (except for a single ``:`` for an empty path)
    * :named_proxy

    These are enabled (rather, displayed) by passing ``True`` in
    ``with_val``, ``with_eval``, ``with_path``, and ``with_proxy``,
    respectively. The latter defaults to `False`, all others to `True`.

    Old behavior (hidden in new mode): Adds separate
    ``-v``/``--var``, ``-e``/``--eval``, -p``/``--path, and
    ``-P``/``--proxy`` arguments.

    Use `attr_args(with_path=False)` to skip path arguments. Ditto for
    `with_eval`.

    Use ``with_combined=False`` to get the old behavior.
    Use ``with_combined=LETTER`` to change the default from ``-s``.

    Old-behavior long options are availabile, but hidden, in "new" mode.
    """

    def _proc(proc):
        if with_combined:
            ht = []
            if with_var:
                ht.append("~str")
            if with_eval:
                ht.append("=value")
            if with_path:
                ht.append(".path")
            if with_proxy:
                ht.append(":name")
            ht = "|".join(ht)

            args = ("--set",) + (
                ("-" + ("s" if isinstance(with_combined, bool) else with_combined),)
                if with_combined
                else ()
            )
            proc = click.option(
                *args,
                "set_",
                nargs=2,
                type=(P, str),
                multiple=True,
                help=f"{par_name} (name {ht})",
                hidden=not with_eval or not with_combined,
            )(proc)

        args = ("--path",) + (("-p",) if with_path else ())
        proc = click.option(
            *args,
            "path_",
            nargs=2,
            type=(P, P),
            multiple=True,
            help=f"{par_name} (name value), as path",
            hidden=not with_path or not with_combined,
        )(proc)

        args = ("--eval",) + (("-e",) if with_eval else ())
        proc = click.option(
            *args,
            "eval_",
            nargs=2,
            type=(P, str),
            multiple=True,
            help=f"{par_name} (name value), evaluated",
            hidden=not with_eval or not with_combined,
        )(proc)

        args = ("--var",) + (("-v",) if with_var else ())
        proc = click.option(
            *args,
            "vars_",
            nargs=2,
            type=(P, str),
            multiple=True,
            help=f"{par_name} (name value)",
            hidden=not with_var or not with_combined,
        )(proc)

        args = ("--proxy",) + (("-P",) if with_proxy else ())
        proc = click.option(
            *args,
            "proxy_",
            nargs=2,
            type=(P, str),
            multiple=True,
            help="Remote proxy (name value)",
            hidden=not with_proxy or not with_combined,
        )(proc)

        return proc

    if proc is None:
        return _proc
    else:
        return _proc(proc)


def process_args(val, set_=(), vars_=(), eval_=(), path_=(), proxy_=(), no_path=False, vs=None):
    """
    process ``set_``/``vars_``/``eval_``/``path_``/``proxy_`` args.

    Arguments:
        val: dict to modify
        set_, vars_, eval_, path_, proxy_: via `attr_args`
        vs: if given: set of vars
    Returns:
        the new value.
    """
    n = 0
    # otherwise these are assumes to be empty tuples.
    if isinstance(vars_, Mapping):
        set_ = set_.items()
    if isinstance(vars_, Mapping):
        vars_ = vars_.items()
    if isinstance(eval_, Mapping):
        eval_ = eval_.items()
    if isinstance(path_, Mapping):
        path_ = path_.items()
    if isinstance(proxy_, Mapping):
        proxy_ = proxy_.items()

    def data():
        for k, v in set_:
            if v[0] == "~":
                v = v[1:]
            elif v == "=-":
                v = NotGiven
            elif v == "=t":
                v = True
            elif v == "=f":
                v = False
            elif v == "=n":
                v = None
            elif v[0] == "=":
                v = eval(v[1:])  # pylint: disable=W0631
            elif v == ":":
                v = P(":")
            elif v[0] == ".":
                v = P(v[1:])
            elif v[0] == ":":
                v = Proxy(v[1:])
            else:
                try:
                    v = int(v)
                except ValueError:
                    try:
                        v = float(v)
                    except ValueError:
                        try:
                            if "." in v:
                                v = P(v)
                            else:
                                raise ValueError("noPath")
                        except ValueError:
                            pass  # leave it as a string
            yield k, v
        for k, v in vars_:
            yield k, v
        for k, v in eval_:
            # ruff:noqa:PLW2901 # var overwritten
            if v == "-":
                v = NotGiven
            elif v == "/":  # pylint: disable=W0631
                if vs is None:
                    raise click.BadOptionUsage(
                        option_name=k,
                        message="A slash value doesn't work here.",
                    )
                v = NoneType
            else:
                v = eval(v)  # pylint: disable=W0631
            yield k, v
        for k, v in path_:
            v = P(v)
            if no_path:
                v = tuple(v)
            yield k, v
        if proxy_:
            if Proxy is None:
                raise ImportError("msgpack")
            for k, v in proxy_:
                v = Proxy(v)
                yield k, v

    if set_:
        dd = data()
    else:
        dd = [(len(k), k, v) for k, v in data()]
        dd.sort()
        dd = [(k, v) for _l, k, v in dd]

    for k, v in dd:
        if isinstance(k, str):
            k = P(k)
        if not len(k):
            if vs is not None:
                raise click.BadOptionUsage(
                    option_name=k,
                    message="You can't use empty paths here.",
                )
            if n:
                raise click.BadOptionUsage(
                    option_name=k,
                    message="Setting a single value conflicts.",
                )
            val = v
            n = -1
        elif n < 0:
            raise click.BadOptionUsage(option_name=k, message="Setting a single value conflicts.")
        else:
            if not isinstance(val, Mapping):
                val = attrdict()
            if vs is not None:
                vs.add(str(k))
            if v is NotGiven:
                val = attrdict._delete(val, k)  # pylint: disable=protected-access
            elif v is NoneType:
                val = attrdict._delete(val, k)  # pylint: disable=protected-access
                vs.discard(str(k))
            else:
                val = attrdict._update(val, k, v)  # pylint: disable=protected-access
            n += 1
    return val


def read_cfg(name, path=None):
    """
    Read a YAML config file, either from the specified path
    or from a couple of default paths.
    """
    cfg = None

    def _cfg(path):
        nonlocal cfg
        if cfg is not None:
            return
        if path is False:
            return
        if os.path.exists(path):
            try:
                with open(path) as cf:
                    cfg = yload(cf, attr=True)
            except PermissionError:
                pass

    if name is not None and cfg is not False:
        if path is not None:
            _cfg(path)
        else:
            _cfg(os.path.expanduser(f"~/.config/{name}.cfg"))
            _cfg(os.path.expanduser(f"~/.{name}.cfg"))
            _cfg(f"/etc/{name}/{name}.cfg")
            _cfg(f"/etc/{name}.cfg")

    return cfg


def load_ext(name, *attr, err=False):
    """
    Load a module
    """
    path = name.split(".")
    path.extend(attr[:-1])
    dp = ".".join(path)
    ".".join(path[:-1])
    try:
        mod = importlib.import_module(dp)
    except ModuleNotFoundError as exc:
        if err and not exc.name.endswith("._main"):
            logger.debug("Err %s: %r", dp, exc, exc_info=exc)
        return None
    except FileNotFoundError:
        if err:
            raise
        return None
    else:
        if attr:
            try:
                mod = getattr(mod, attr[-1])
            except AttributeError:
                logger.debug("Err %s.%s", dp, attr[-1])
                return None
        return mod


def load_cfg(name):
    """
    Load a module's configuration
    """
    ensure_cfg(name)
    return CFG


def _namespaces(name):
    import pkgutil  # pylint: disable=import-outside-toplevel

    try:
        ext = importlib.import_module(name)
    except ModuleNotFoundError:
        logger.debug("No NS: %s", name)
        return ()
    try:
        p = ext.__path__
    except AttributeError:
        p = (str(FSPath(ext.__file__).parent),)
    logger.debug("NS: %s %s", name, p)
    return pkgutil.iter_modules(p, ext.__name__ + ".")


_ext_cache = defaultdict(dict)


def _cache_ext(ext_name, pkg_only):
    """List external modules

    Yields (name,path) tuples.

    TODO: This is not zip safe.
    """
    for finder, name, ispkg in _namespaces(ext_name):
        if pkg_only and not ispkg:
            logger.debug("ExtNoC %s", name)
            continue
        logger.debug("ExtC %s", name)
        x = name.rsplit(".", 1)[-1]
        f = FSPath(finder.path) / x
        _ext_cache[ext_name][x] = f


def list_ext(name, func=None, pkg_only=True):
    """List external modules

    Yields (name,path) tuples.

    TODO: This is not zip safe.
    """
    logger.debug("List Ext %s (%s)", name, func)
    if name not in _ext_cache:
        with suppress(ModuleNotFoundError):
            _cache_ext(name, pkg_only)
    if func is None:
        for a, b in _ext_cache[name].items():
            logger.debug("Found %s %s", a, b)
            yield a, b
        return

    for x, f in _ext_cache[name].items():
        if (f / ".no_load").is_file():
            logger.debug("Skip %s", f)
            continue
        fn = f / (func + ".py")
        if not fn.is_file():
            fn = f / func / "__init__.py"
            if not fn.is_file():
                # XXX this might be a namespace
                logger.debug("No file: %s/%s", f, func)
                continue
        logger.debug("Found2 %s %s", x, f)
        yield (x, f)


def load_subgroup(
    _fn=None,
    prefix=None,
    sub_pre=None,
    sub_post=None,
    ext_pre=None,
    ext_post=None,
    **kw,
) -> Command:
    """
    A decorator like click.group, enabling loading of subcommands

    Internal extensions are loaded as ``{sub_pre}.*.{sub_post}``.
    External extensions are loaded as ``{ext_pre}.*.{ext_post}``.

    All other arguments are forwarded to `click.command`.
    """

    def _ext(fn, **kw):
        return click.command(**kw)(fn)

    kw["cls"] = partial(
        kw.get("cls", Loader),
        _util_sub_pre=sub_pre or this_load.get() or prefix,
        _util_sub_post=sub_post or (None if prefix is None else "cli"),
        _util_ext_pre=ext_pre or prefix,
        _util_ext_post=ext_post or (None if prefix is None else "_main.cli"),
    )

    if _fn is None:
        return partial(_ext, **kw)
    else:
        return _ext(_fn, **kw)


class Loader(click.Group):
    """
    A `click.group` that loads additional commands from subfolders and/or extensions.

    Subfolders: set _util_sub_pre to your module's name.
        This works with namespace packages.
        E.g. "distkv.command" loads "distkv.command.*.cli".

    Extensions: set _util_ext_pre to the extension basename.
        Set _util_ext_post to the name of the extension.

        E.g. "distkv_ext"+"client" loads "distkv_ext.*.client.cli".

    Both work in parallel.

    Caller:

        from moat.util import Loader
        from functools import partial

        @click.command(cls=partial(Loader,_util_sub_post='command'))
        async def cmd()
            print("I am the main program")

    Sub-Command Usage (``main`` is defined for you), e.g. in ``command/subcmd.py``::

        from moat.util import Loader
        from functools import partial

        @main.command / group()
        async def cmd(self):
            print("I am", self.name)  # prints "subcmd"
    """

    # ruff:noqa:SLF001

    def __init__(
        self,
        *,
        _util_sub_pre=None,
        _util_sub_post=None,
        _util_ext_pre=None,
        _util_ext_post=None,
        **kw,
    ):
        logger.debug(
            "* Load: %s.*.%s / %s.*.%s",
            _util_sub_pre,
            _util_sub_post,
            _util_ext_pre,
            _util_ext_post,
        )
        if _util_sub_pre is not None:
            self._util_sub_pre = _util_sub_pre
        if _util_sub_post is not None:
            self._util_sub_post = _util_sub_post
        if _util_ext_pre is not None:
            self._util_ext_pre = _util_ext_pre
        if _util_ext_post is not None:
            self._util_ext_post = _util_ext_post
        super().__init__(**kw)

    def get_sub_ext(self, ctx):
        """Fetch extension variables"""
        sub_pre = getattr(
            # pylint: disable=protected-access
            self,
            "_util_sub_pre",
            ctx.obj._util_sub_pre,
        )
        sub_post = getattr(
            # pylint: disable=protected-access
            self,
            "_util_sub_post",
            ctx.obj._util_sub_post,
        )
        ext_pre = getattr(
            # pylint: disable=protected-access
            self,
            "_util_ext_pre",
            ctx.obj._util_ext_pre,
        )
        ext_post = getattr(
            # pylint: disable=protected-access
            self,
            "_util_ext_post",
            ctx.obj._util_ext_post,
        )

        if sub_pre is None:
            sub_post = None
        elif sub_post is None:
            sub_pre = ("cli",)
        elif isinstance(sub_post, str):
            sub_post = sub_post.split(".")

        if ext_pre is None:
            ext_post = None
        elif ext_post is None:
            ext_pre = None
        elif isinstance(ext_post, str):
            ext_post = ext_post.split(".")
            if len(ext_post) == 1:
                ext_post.append("cli")

        return sub_pre, sub_post, ext_pre, ext_post

    def list_commands(self, ctx):
        "show subpackages"
        rv = super().list_commands(ctx)
        sub_pre, sub_post, ext_pre, ext_post = self.get_sub_ext(ctx)
        logger.debug("* List: %s.*.%s / %s.*.%s", sub_pre, sub_post, ext_pre, ext_post)

        if sub_pre:
            logger.debug("Adding sub %s",sub_pre)
            for _finder, name, _ispkg in _namespaces(sub_pre):
                # ruff:noqa:PLW2901 # var overwritten
                name = name.rsplit(".", 1)[1]
                if name[0] == "_":
                    continue
                if load_ext(sub_pre, name, *sub_post, err=ctx.obj.debug_loader):
                    rv.append(name)

        if ext_pre:
            logger.debug("Adding ext %s",ext_pre)
            for n, _ in list_ext(ext_pre):
                if load_ext(ext_pre, n, *ext_post, err=ctx.obj.debug_loader):
                    rv.append(n)
        rv.sort()
        logger.debug("List: %r", rv)
        return rv

    def get_command(self, ctx, cmd_name):
        "add subpackages"
        command = super().get_command(ctx, cmd_name)

        sub_pre, sub_post, ext_pre, ext_post = self.get_sub_ext(ctx)

        if command is None and ext_pre is not None:
            command = load_ext(ext_pre, cmd_name, *ext_post)
            if command is not None:
                cf = load_cfg(f"{ext_pre}.{cmd_name}")
                merge(ctx.obj.cfg, cf, replace=False)

        if command is None:
            if sub_pre is None:
                return None
            command = load_ext(sub_pre, cmd_name, *sub_post)
            if command is not None:
                cf = load_cfg(f"{sub_pre}.{cmd_name}")
                merge(ctx.obj.cfg, cf, replace=False)

        if command is None:
            # raise click.UsageError(f"No such subcommand: {cmd_name}")
            return None
        command.__name__ = command.name = cmd_name
        return command


class MainLoader(Loader):
    """
    A special loader that runs the main setup code even if there's a
    subcommand with "--help".
    """

    async def invoke(self, ctx):
        if not getattr(ctx, "_moat_invoked", False):
            await ctx.invoke(self.callback, **ctx.params)
        return await super().invoke(ctx)


#
# There are two ways this can start up.
# (a) `main_` is the "real" main function. It sets up the Click environment and then
#     starts anyio and runs the function body, which calls `wrap_main`
#     synchronously to set up our object.
#
# (b) `wrap_main` is used as a wrapper, used mainly for testing. It sets up the context
#     and then returns "main_.main()", which is an awaitable, thus
#     `wrap_main` acts as an async function.


@load_subgroup(
    cls=MainLoader,
    add_help_option=False,
    invoke_without_command=True,
)  # , __file__, "command"))
@click.option("-V", "--verbose", count=True, help="Be more verbose. Can be used multiple times.")
@click.option("-L", "--debug-loader", is_flag=True, help="Debug submodule loading.")
@click.option("-Q", "--quiet", count=True, help="Be less verbose. Opposite of '--verbose'.")
@click.option("-D", "--debug", count=True, help="Enable debug speed-ups (smaller keys etc).")
@click.option(
    "-l",
    "--log",
    multiple=True,
    help="Adjust log level. Example: '--log asyncactor=DEBUG'.",
)
@click.option(
    "-c",
    "--cfg",
    type=click.Path("r"),
    default=None,
    help="Configuration file (YAML).",
    multiple=True,
)
@click.option(
    "-h",
    "-?",
    "--help",
    is_flag=True,
    help="Show help. Subcommands only understand '--help'.",
)
@attr_args(par_name="Config item")
@click.pass_context
async def main_(ctx, verbose, quiet, help=False, **kv):  # pylint: disable=redefined-builtin
    """
    This is the main command. (You might want to override this text.)

    You need to add a subcommand for this to do anything.
    """
    ctx.allow_interspersed_args = True

    # The above `MainLoader.invoke` call causes this code to be called
    # twice instead of never.
    if hasattr(ctx, "_moat_invoked"):
        return
    ctx._moat_invoked = True  # pylint: disable=protected-access
    wrap_main(ctx=ctx, verbose=max(0, 1 + verbose - quiet), **kv)
    if help or ctx.invoked_subcommand is None and not ctx.protected_args:
        print(ctx.get_help())
        ctx.exit()


def wrap_main(  # pylint: disable=redefined-builtin,inconsistent-return-statements
    main=main_,
    *,
    set_=(),
    vars_=(),
    eval_=(),
    path_=(),
    proxy_=(),
    name=None,
    sub_pre=None,
    sub_post=None,
    ext_pre=None,
    ext_post=None,
    cfg=None,
    CFG=None,
    args=None,
    wrap=False,
    verbose=1,
    debug=0,
    debug_loader=False,
    log=(),
    ctx=None,
    help=None,
) -> Awaitable:
    """
    The main command entry point, when testing.

    main: special main function, defaults to moat.util.main_
    name: command name, defaults to {main}'s toplevel module name.
    {sub,ext}_{pre,post}: commands to load in submodules or extensions.

    cfg: configuration file(s), default: various locations based on {name}, False=don't load
    CFG: default configuration (dir or file), relative to caller
         Default: load from name._config

    wrap: Flag: this is a subcommand. Don't set up logging, return the awaitable.
    args: Argument list if called from a test, `None` otherwise.
    help: Help text of your code.

    Internal extensions are loaded as ``{sub_pre}.*.{sub_post}``.
    External extensions are loaded as ``{ext_pre}.*.{ext_post}``.

    cfg.moat may contain values for {sub,ext}_{pre,post}.
    """

    obj = getattr(ctx, "obj", None)
    if obj is None:
        obj = attrdict()

    opts = obj.get("moat", None)
    if opts is None:
        obj.moat = opts = attrdict()

    if sub_pre is None:
        sub_pre = opts.get("sub_pre", None)
    else:
        opts["sub_pre"] = sub_pre

    if sub_post is None:
        sub_post = opts.get("sub_post", None)
    else:
        opts["sub_post"] = sub_post

    if ext_pre is None:
        ext_pre = opts.get("ext_pre", None)
    else:
        opts["ext_pre"] = ext_pre

    if ext_post is None:
        ext_post = opts.get("ext_post", None)
    else:
        opts["ext_post"] = ext_post

    if name is None:
        name = opts.get("name", "moat")
    else:
        opts["name"] = name

    if sub_pre is True:
        import inspect  # pylint: disable=import-outside-toplevel

        sub_pre = inspect.currentframe().f_back.f_globals["__package__"]
    elif sub_pre is None:
        sub_pre = name
    if sub_post is None:
        sub_post = "_main.cli"

    if main is None:
        if help is not None:
            raise RuntimeError("You can't set the help text this way")
    else:
        main.context_settings["obj"] = obj
        if help is not None:
            main.help = help

    obj._util_sub_pre = sub_pre  # pylint: disable=protected-access
    obj._util_sub_post = sub_post  # pylint: disable=protected-access
    obj._util_ext_pre = ext_pre  # pylint: disable=protected-access
    obj._util_ext_post = ext_post  # pylint: disable=protected-access

    if CFG is None:
        CFG = opts.get("CFG")

    if isinstance(CFG, str):
        p = FSPath(CFG)
        if not p.is_absolute():
            p = FSPath((main or main_).__file__).parent / p
        with open(p) as cfgf:
            CFG = yload(cfgf, attr=True)
    elif CFG is None:
        CFG = obj.get("CFG", None)
        if CFG is None:
            CFG = load_cfg(name)

    obj.stdout = CFG.get("_stdout", sys.stdout)  # used for testing
    obj.CFG = CFG

    if not cfg:
        cfg = to_attrdict(read_cfg(name, None))
    elif isinstance(cfg, (list, tuple)):
        cf = {}
        for fn in cfg:
            merge(cf, read_cfg(name, fn), replace=True)
        cfg = to_attrdict(cf)
    else:
        cfg = to_attrdict(read_cfg(name, cfg))

    if cfg:
        merge(cfg, obj.CFG, replace=False)
    else:
        cfg = CFG

    if (fn := os.environ.get("MOAT_CFG")) is not None:
        merge(cfg, read_cfg(name, fn), replace=False)

    obj.cfg = cfg = to_attrdict(cfg)

    obj.debug = verbose
    obj.DEBUG = debug

    # Don't forget to import toplevel config files
    import moat

    try:
        p = moat.__path__
    except AttributeError:
        p = (str(FSPath(ext.__file__).parent),)
    for fp in p:
        try:
            with (FSPath(fp) / "_cfg.yaml").open("r") as f:
                merge(cfg, yload(f, attr=True), replace=False)
        except FileNotFoundError:
            pass

    obj.cfg = process_args(
        obj.cfg,
        set_=set_,
        vars_=vars_,
        eval_=eval_,
        path_=path_,
        proxy_=proxy_,
    )

    if wrap:
        pass
    elif hasattr(logging.root, "_MoaT"):
        logging.debug("Logging already set up")
    else:
        # Configure logging. This is a somewhat arcane art.
        lcfg = obj.cfg.setdefault("logging", {})
        lcfg.setdefault("version", 1)
        lcfg.setdefault("root", {})["level"] = (
            "DEBUG"
            if verbose > 2
            else "INFO"
            if verbose > 1
            else "WARNING"
            if verbose
            else "ERROR"
        )
        for k in log:
            k, v = k.split("=")
            lcfg["loggers"].setdefault(k, {})["level"] = v
        logging.config.dictConfig(lcfg)

        logging.basicConfig = _no_config
        logging.config.dictConfig = _no_config
        logging.config.fileConfig = _no_config

        logging.captureWarnings(verbose > 0)
        logger.disabled = False
        if debug_loader:
            logger.level = logging.DEBUG
            for p in sys.path:
                logger.debug("Path: %s", p)
        logging.root._MoaT = True

    obj.logger = logging.getLogger(name)
    obj.debug_loader = debug_loader

    try:
        # pylint: disable=no-value-for-parameter,unexpected-keyword-arg
        # NOTE this return an awaitable
        if ctx is not None:
            ctx.obj = obj
        elif main is not None:
            if wrap:
                main = main.main
            with ungroup():
                return main(args=args, standalone_mode=False, obj=obj)

    except click.exceptions.MissingParameter as exc:
        print(
            f"You need to provide an argument {exc.param.name.upper()!r}.\n",
            file=sys.stderr,
        )
        print(exc.cmd.get_help(exc.ctx), file=sys.stderr)
        sys.exit(2)
    except click.exceptions.UsageError as exc:
        try:
            s = str(exc)
        except TypeError:
            logger.exception("??", exc_info=exc)
        else:
            print(s, file=sys.stderr)
        sys.exit(2)
    except click.exceptions.Abort:
        print("Aborted.", file=sys.stderr)


def _ng(type_):
    @wraps(type_)
    def gen(data):
        if data is NotGiven:
            return data
        return type_(data)

    return gen


def option_ng(*a, type=str, **kw):
    return click.option(*a, **kw, type=_ng(type), default=NotGiven)



"""
This module contains various helper functions and classes for dictionaries.
"""

from __future__ import annotations

from collections.abc import Mapping
from copy import deepcopy

from . import NotGiven

__all__ = ["combine_dict", "drop_dict", "attrdict", "to_attrdict"]


def combine_dict(*d, cls=dict, deep=False) -> dict:
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.

    This operation is recursive and non-destructive. If `deep` is set, the
    result always is a deep copy.

    A value of `NotGiven` causes an entry to be skipped.

    TODO: arrays are not merged.

    Args:
      cls (type): a class to instantiate the result with. Default: dict.
        Often used: :class:`attrdict`.
      deep (bool): if set, always copy.
    """
    res = cls()
    keys = {}
    if not d:
        return res

    if len(d) == 1 and deep and not isinstance(d[0], Mapping):
        if deep and isinstance(d[0], (list, tuple)):
            return deepcopy(d[0])
        else:
            return d[0]

    for kv in d:
        if kv is None:
            continue
        for k, v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)

    for k, v in keys.items():
        if v[0] is NotGiven:
            pass
        elif len(v) == 1 and not deep:
            res[k] = v[0]
        elif not isinstance(v[0], Mapping):
            for vv in v[1:]:
                assert vv is NotGiven or not isinstance(vv, Mapping)
            if deep and isinstance(v[0], (list, tuple)):
                res[k] = deepcopy(v[0])
            else:
                res[k] = v[0]
        else:
            res[k] = combine_dict(*v, cls=cls)

    return res


def drop_dict(data: Mapping, drop: tuple[str | tuple[str]]) -> Mapping:
    """
    Helper to remove some entries from a mapping hierarchy

    Returns a new mapping. The original is not changed.
    """
    data = data.copy()
    for d in drop:
        # ruff:noqa:PLW2901 # var overwritten
        vv = data
        if isinstance(d, (tuple, list)):
            for dd in d[:-1]:
                vv = vv[dd] = vv[dd].copy()
            d = d[-1]
        del vv[d]
    return data


class attrdict(dict):
    """
    A dictionary which can be accessed via attributes, for convenience.

    This also supports updating path accessors.
    """

    def __getattr__(self, a):
        if a.startswith("_"):
            return object.__getattribute__(self, a)
        try:
            return self[a]
        except KeyError:
            raise AttributeError(a) from None

    def __setattr__(self, a, b):
        if a.startswith("_"):
            super().__setattr__(a, b)
        else:
            self[a] = b

    def __delattr__(self, a):
        try:
            del self[a]
        except KeyError:
            raise AttributeError(a) from None

    def _get(self, path, skip_empty=True, default=NotGiven):
        """
        Get a node's value and access the dict items beneath it.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        val = self
        for p in path:
            if val is None:
                return None
            if skip_empty and not p:
                continue
            val = val.get(p, NotGiven)
            if val is NotGiven:
                if default is NotGiven:
                    raise KeyError(path)
                return default
        return val

    def _update(self, path, value=None, skip_empty=False):
        """
        Set some sub-item's value, possibly merging dicts.
        Items set to 'NotGiven' are deleted.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        if skip_empty:
            path = [p for p in path if p != ""]
        val = type(self)(**self)
        v = val
        if not path:
            if isinstance(value, Mapping):
                return combine_dict(value, val, cls=type(self))
            else:
                return value

        for p in path[:-1]:
            try:
                w = v[p]
            except KeyError:
                w = type(v)()
            else:
                # copy
                w = attrdict() if w is None else type(w)(w)
            v[p] = w
            v = w
        px = path[-1]
        if value is NotGiven:
            v.pop(px, None)
        elif isinstance(v, list) and px is None:
            v.append(value)
        elif not isinstance(value, Mapping):
            v[px] = value
        elif px in v:
            v[px] = combine_dict(value, v[px], cls=type(self))
        else:
            v[px] = value

        return val

    def _delete(self, path, skip_empty=True):
        """
        Remove some sub-item's value, possibly removing now-empty intermediate
        dicts.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise TypeError(f"Must be a Path/list, not {path!r}")
        if skip_empty:
            path = [p for p in path if p]
        val = type(self)(**self)
        v = val
        vc = []
        for p in path[:-1]:
            vc.append(v)
            try:
                w = v[p]
            except KeyError:
                return self
            w = type(w)(**w)
            v[p] = w
            v = w
        vc.append(v)
        while path:
            v = vc.pop()
            v.pop(path.pop(), None)
            if v:
                break
        return val


def to_attrdict(d) -> attrdict:
    """
    Return a hierarchy with all dicts converted to attrdicts.
    """
    if isinstance(d, dict):
        return attrdict((k, to_attrdict(v)) for k, v in d.items())
    if isinstance(d, (tuple, list)):
        return [to_attrdict(v) for v in d]
    return d



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import sys
from functools import partial
from types import ModuleType

__all__ = ["make_proc", "Module", "make_module"]


def _call_proc(code, variables, *a, **kw):
    v = variables[len(a) :]
    if v:
        a = list(a)
        for k in v:
            a.append(kw.pop(k, None))
    eval(code, kw)  # pylint: disable=eval-used
    code = kw["_proc"]
    return code(*a)


def make_proc(code, variables, path, *, use_async=False):  # pylint: disable=redefined-builtin
    """Compile this code block to a procedure.

    Args:
        code: the code block to execute. Text, will be indented.
        vars: variable names to pass into the code
        path: the location where the code is stored
        use_async: False if sync code, True if async, None if in thread
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    hdr = f"""\
def _proc({",".join(variables)}):
    """

    if use_async:
        hdr = "async " + hdr
    code = hdr + code.replace("\n", "\n    ")
    code = compile(code, str(path), "exec")

    return partial(_call_proc, code, variables)


class Module(ModuleType):
    """A dynamically-loaded module.

    TODO.
    """

    def __repr__(self):
        return f"<Module {self.__class__.__name__}%s>"


def make_module(code, path):
    """Compile this code block to something module-ish.

    Args:
        code: the code block to execute
        path: the location where the code is / shall be stored
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    name = ".".join(str(x) for x in path)
    code = compile(code, name, "exec")
    m = sys.modules.get(name, None)
    if m is None:
        m = ModuleType(name)
    eval(code, m.__dict__)  # pylint: disable=eval-used
    sys.modules[name] = m
    return m



"""
Exception handling helpers

This code is *duplicated* in moat.micro:_embed/lib/moat/util/exc.py
"""

from __future__ import annotations
import os
from anyio import get_cancelled_exc_class
from sniffio import AsyncLibraryNotFoundError

__all__ = [
    "exc_iter",
    "ungroup",
    "ExpectedError",
    "ExpKeyError",
    "ExpAttrError",
]


class ExpectedError(Exception):
    """
    An error that shouldn't elicit a traceback
    """

    def __init__(self, exc):
        self.exc = exc


class ExpKeyError(KeyError, ExpectedError):
    "unreported key error"

    pass


class ExpAttrError(AttributeError, ExpectedError):
    "unreported key error"

    pass


def exc_iter(exc):
    """
    iterate over all non-exceptiongroup parts of an exception(group)
    """
    if isinstance(exc, BaseExceptionGroup):
        for e in exc.exceptions:
            yield from exc_iter(e)
    else:
        yield exc


class ungroup:
    """
    A sync+async context manager that unwraps single-element
    exception groups.
    """

    def __call__(self):
        "Singleton. Returns itself."
        return self

    @staticmethod
    def one(e):
        "convert the exceptiongroup @e to a single exception"
        if not isinstance(e, BaseExceptionGroup):
            return e

        try:
            Cancel = get_cancelled_exc_class()
        except AsyncLibraryNotFoundError:
            pass
        else:
            c, e = e.split(Cancel)
            if not e:
                e = c

        while isinstance(e, BaseExceptionGroup):
            if len(e.exceptions) != 1:
                break
            e = e.exceptions[0]
        return e

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, c, e, t):
        if e is None:
            return
        if "MOAT_TB" in os.environ:
            return
        e = self.one(e)
        raise e from None

    async def __aexit__(self, c, e, t):
        return self.__exit__(c, e, t)


ungroup = ungroup()



"""
An overly-simple CBOR packer/unpacker.
"""

# Original Copyright 2014-2015 Brian Olson
# Apache 2.0 license
# http://docs.ros.org/en/noetic/api/rosbridge_library/html/cbor_8py_source.html
# rather heavily modified

from __future__ import annotations

import datetime as dt
import re
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)

# Typing
from moat.lib.codec.cbor import CBOR_TAG_CBOR_FILEHEADER, CBOR_TAG_CBOR_LEADER, Codec, Tag

from ._cbor import std_ext, StdCBOR
import contextlib

__all__ = ["std_ext", "StdCBOR", "gen_start", "gen_stop"]

CBOR_TAG_MOAT_FILE_ID = 1299145044  # 'MoaT'
CBOR_TAG_MOAT_FILE_END = 1298493254  # 'MeoF'
CBOR_TAG_MOAT_CHANGE = 1298360423  # 'Mchg'

Codec = StdCBOR


def gen_start(text: str, /, **kw) -> Tag:
    """
    Generate a MoaT file start tag
    """
    if len(text) > 255:
        raise ValueError("Description too long")
    # add padding if too short
    text += " " * (24 - len(text))

    return Tag(CBOR_TAG_CBOR_LEADER, Tag(CBOR_TAG_MOAT_FILE_ID, (text, kw)))


def gen_stop(**kw) -> Tag:
    """
    Generate a MoaT file stop tag
    """
    return Tag(CBOR_TAG_MOAT_FILE_END, kw)


def gen_change(**kw) -> Tag:
    """
    Generate a MoaT file stop tag
    """
    return Tag(CBOR_TAG_MOAT_CHANGE, kw)


@std_ext.encoder(1, dt.datetime)
def _enc_datetime_ts(codec, value):
    codec  # noqa:B018
    if not value.tzinfo:
        raise ValueError(f"naive datetime {value!r}")

    return value.timestamp()


# @std_ext.encoder(0, dt.datetime)
def _enc_datetime_str(codec, value):
    codec  # noqa:B018

    if not value.tzinfo:
        raise ValueError(f"naive datetime {value!r}")

    return value.isoformat().replace("+00:00", "Z")


_timestamp_re = re.compile(
    r"^(\d{4})-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)(?:\.(\d{1,6})\d*)?(?:Z|([+-])(\d\d):(\d\d))$",
)


@std_ext.decoder(0)
def _dec_datetime_string(codec, value) -> dt.datetime:
    codec  # noqa:B018

    match = _timestamp_re.match(value)
    if match:
        (
            year,
            month,
            day,
            hour,
            minute,
            second,
            secfrac,
            offset_sign,
            offset_h,
            offset_m,
        ) = match.groups()
        microsecond = 0 if secfrac is None else int(f"{secfrac:<06}")

        if offset_h:
            sign = -1 if offset_sign == "-" else 1
            hours = int(offset_h) * sign
            minutes = int(offset_m) * sign
            tz = dt.timezone(dt.timedelta(hours=hours, minutes=minutes))
        else:
            tz = dt.UTC

        return dt.datetime(
            int(year),
            int(month),
            int(day),
            int(hour),
            int(minute),
            int(second),
            microsecond,
            tz,
        )
    else:
        raise ValueError(f"invalid datetime string: {value!r}")


@std_ext.decoder(1)
def _dec_ts(codec, val):
    codec  # noqa:B018
    return dt.datetime.fromtimestamp(val, dt.UTC)


@std_ext.decoder(2)
def _dec_bigp(codec, val):
    codec  # noqa:B018
    return int.from_bytes(val, "big")


@std_ext.decoder(3)
def _dec_bign(codec, val):
    codec  # noqa:B018
    return -1 - int.from_bytes(val, "big")


def _pad(buf, n):
    buf = bytes(buf)  # SIGH
    if (lb := len(buf)) == n:
        return buf
    return buf + b"\x00" * (n - lb)


@std_ext.decoder(52)
def _dec_ip4address(codec, buf) -> IPv4Address | IPv4Network | IPv4Interface:
    codec  # noqa:B018
    from ipaddress import IPv4Address, IPv4Interface, IPv4Network

    if isinstance(buf, (bytes, bytearray, memoryview)):
        return IPv4Address(_pad(buf, 4))
    elif len(buf) != 2:
        pass
    elif isinstance(buf[1], int):
        return IPv4Interface((_pad(buf[0], 4), buf[1]))
    else:
        return IPv4Network((_pad(buf[1], 4), buf[0]))

    raise ValueError(f"invalid ipaddress value {buf!r}")


@std_ext.decoder(54)
def _dec_ip6address(codec, buf) -> IPv6Address | IPv6Network | IPv6Interface:
    codec  # noqa:B018
    from ipaddress import IPv6Address, IPv6Interface, IPv6Network

    if isinstance(buf, (bytes, bytearray, memoryview)):
        return IPv6Address(_pad(buf, 16))
    elif len(buf) != 2:
        pass
    elif isinstance(buf[1], int):
        return IPv6Interface((_pad(buf[0], 16), buf[1]))
    else:
        return IPv6Network((_pad(buf[1], 16), buf[0]))

    raise ValueError(f"invalid ipaddress value {buf!r}")


@std_ext.encoder(52, IPv4Address)
@std_ext.encoder(54, IPv6Address)
def _pack_ip(codec, adr):
    codec  # noqa:B018
    return adr.packed.rstrip(b"\x00")


@std_ext.encoder(52, IPv4Network)
@std_ext.encoder(54, IPv6Network)
def _pack_ipnet(codec, adr):
    codec  # noqa:B018
    return (adr.prefixlen, _pack_ip(codec, adr.network_address))


@std_ext.encoder(52, IPv4Interface)
@std_ext.encoder(54, IPv6Interface)
def _pack_ipintf(codec, adr):
    codec  # noqa:B018
    return (_pack_ip(codec, adr), adr.network.prefixlen)


@std_ext.decoder(260)
def _dec_old_ipaddress(codec, buf) -> IPv4Address | IPv6Address | Tag:
    codec  # noqa:B018
    if isinstance(buf, (bytes, bytearray, memoryview)) or len(buf) not in (4, 6, 16):
        if len(buf) == 4:
            return IPv4Address(bytes(buf))
        elif len(buf) == 16:
            return IPv6Address(bytes(buf))
        elif len(buf) == 6:
            # MAC address
            return Tag(260, buf)

    raise ValueError(f"invalid ipaddress value {buf!r}")


@std_ext.decoder(261)
def _dec_old_ipnetwork(codec, buf) -> IPv4Network | IPv6Network:
    codec  # noqa:B018
    if isinstance(buf, dict) and len(buf) == 1:
        mask, buf = next(iter(buf.items()))
        if len(buf) == 4:
            return IPv4Network((bytes(buf), mask))
        elif len(buf) == 16:
            return IPv6Network((bytes(buf), mask))

    raise ValueError(f"invalid ipnetwork value {buf!r}")


@std_ext.decoder(CBOR_TAG_CBOR_FILEHEADER)
def _dec_file_cbor(codec, val):
    codec  # noqa:B018
    with contextlib.suppress(AttributeError):
        val._cbor_tag = CBOR_TAG_CBOR_FILEHEADER
    return val


@std_ext.decoder(CBOR_TAG_CBOR_LEADER)
def _dec_file_cbor(codec, val):
    codec  # noqa:B018
    with contextlib.suppress(AttributeError):
        val._cbor_tag = CBOR_TAG_CBOR_LEADER
    return val



# noqa: D100  # compatibility with µPy
from __future__ import annotations

import os

rs = os.environ.get("PYTHONHASHSEED", None)
if rs is None:
    import random
else:  # pragma: no cover
    try:
        import trio._core._run as tcr
    except ImportError:
        import random
    else:
        random = tcr._r

# Intended to be unambiguous, no special characters
al_unique = "bcdfghjkmnpqrstvwxyzBCDFGHJKMNPQRSTVWXYZ23456789"

# Lowercase and digits, e.g. for restricted-alphabet labels
al_lower = "abcdefghijklmnopqrstuvwxyz0123456789"

# everything (except for backslash, just to be safe)
al_ascii = bytes(x for x in range(33, 127) if x != 92).decode("ascii")

# lowercase letters only
al_az = "abcdefghijklmnopqrstuvwxyz"

__all__ = ["id2str", "gen_ident", "al_unique", "al_lower", "al_ascii", "al_az"]


def gen_ident(k=10, /, *, alphabet=al_unique):
    """
    Generate a random identifier / password.
    """
    return "".join(random.choices(alphabet, k=k))


def id2str(k, /, *, alphabet=al_unique):
    """
    Given a non-numeric identifier, returns a string representing it.
    """
    assert k > 0
    if k == 0:
        return alphabet[0]
    res = []
    la = len(alphabet)
    while k:
        res.append(alphabet[k%la])
        k //= la
    return ''.join(res)





"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import anyio

__all__ = ["spawn"]


async def spawn(taskgroup, proc, *args, **kw):
    """
    Run a task within this object's task group.

    Returns:
        a cancel scope you can use to stop the task.
    """

    async def _run(proc, args, kw, *, task_status):
        """
        Helper for starting a task within a cancel scope.
        """
        with anyio.CancelScope() as sc:
            task_status.started(sc)
            await proc(*args, **kw)

    return await taskgroup.start(_run, proc, args, kw)



"""
An overly-simple CBOR packer/unpacker.
"""

# Original Copyright 2014-2015 Brian Olson
# Apache 2.0 license
# http://docs.ros.org/en/noetic/api/rosbridge_library/html/cbor_8py_source.html
# rather heavily modified

from __future__ import annotations

# Typing
from moat.lib.codec import Extension, NoCodecError
from moat.lib.codec.cbor import Codec, Tag
from moat.lib.codec.proxy import DProxy, Proxy, name2obj, obj2name, unwrap_obj, wrap_obj, get_proxy

from . import NotGiven
from .path import Path

__all__ = ["std_ext", "StdCBOR"]

std_ext = Extension()


class StdCBOR(Codec):
    """
    CBOR codec with MoaT's standard extensions.
    """

    def __init__(self):
        super().__init__(ext=std_ext)

    def decode(self, data:bytes):
        if data == b'':
            return NotGiven
        return super().decode(data)

Codec = StdCBOR


@std_ext.encoder(27, DProxy)
def _enc_dpr(codec, obj):
    codec  # noqa:B018
    res = [obj.name] + obj.a
    if obj.k or res and isinstance(res[-1], dict):
        res.append(obj.k)
    return res


@std_ext.encoder(32769, Proxy)
def _enc_pr(codec, obj):
    codec  # noqa:B018
    return obj.name


@std_ext.decoder(32769)
def _dec_proxy(codec, val):
    codec  # noqa:B018
    try:
        if isinstance(val, (str, int)):
            return name2obj(val)
        return unwrap_obj(val)
    except KeyError:
        return Proxy(val)


@std_ext.decoder(27)
def _dec_obj(codec, val):
    codec  # noqa:B018
    if isinstance(val[0], Tag):
        if val[0].tag != 32769:
            return Tag(27, val)  # not decodable
        val[0] = val[0].value
    return unwrap_obj(val)


@std_ext.encoder(39, Path)
def _enc_path(codec, val):
    codec  # noqa:B018
    return val.raw


@std_ext.decoder(39)
def _dec_path(codec, val):
    codec  # noqa:B018
    if not isinstance(val, (list, tuple)):
        return Tag(39, val)  # not decodable
    return Path.build(val)


@std_ext.encoder(None, object)
def enc_any(codec, obj):
    codec  # noqa:B018

    try:
        name = obj2name(obj)
    except KeyError:
        pass
    else:
        return 32769, name

    try:
        name = obj2name(type(obj))
    except KeyError:
        pass
    else:
        p = wrap_obj(obj, name=name)
        return 27, p

    if isinstance(obj, Exception):
        # RemoteError, cf. moat.lib.codec.errors
        res = ["_rErr", obj.__class__.__name__]
        try:
            res.extend(obj.args)
        except AttributeError:
            pass
        return 27, res

    return 32769, get_proxy(obj)



"""
This module contains functions dealing with MoaT's Path objects.

MoaT Paths are represented as dot-separated strings. The colon is special.
See "moat util path" / moat.util.Path.__doc__ for details.

Behind the scenes, a Path is an immutable lists with special representation.
This is particularly useful in YAML (MoaT uses a ``!P`` prefix).

They are also marked in msgpack and CBOR.

Joining the string representation of two non-empty paths requires a dot iff
the second part doesn't start with a separator-colon, but you shouldn't
ever want to do that: paths are best processed as Path objects, not
strings.


This module also exports the magic paths "Root", "S_Root", "P_Root" and
"Q_Root". These are contextvars and can be set externally. They are
interpolated into slashed Path representations.

"""

from __future__ import annotations

import ast
import collections.abc
import logging
import re
import warnings
from base64 import b64decode, b64encode
from contextvars import ContextVar
from functools import total_ordering

import simpleeval

from moat.lib.codec.proxy import as_proxy
from . import NotGiven

__all__ = [
    "Path",
    "P",
    "PS",
    "logger_for",
    "PathShortener",
    "PathLongener",
    "path_eval",
    "Root",
    "RootPath",
    "set_root",
]

_PartRE = re.compile("[^:._]+|_|:|\\.")
_RTagRE = re.compile("^:m[^:._]+:$")


def set_root(cfg):
    """
    Utility to force-set the root global.

    Used for testing without a "proper" context.
    """
    from moat.util.path import Root, S_Root,P_Root,Q_Root
    Root.set(cfg.root)


@total_ordering
class Path(collections.abc.Sequence):
    """
    Paths are represented as dot-separated strings. The colon is special.
    Inline (within an element):

    \b
        ::  escapes : colon
        :.  escapes . dot   (dot-path repr only)
        :_  escapes   space (dot-path repr only)
        :|  escapes / slash (slash-path repr only)
        :h  escapes # hash  (slash-path repr only, optional)
        :p  escapes + plus  (slash-path repr only, optional)

    As separator (starts a new element):

    \b
        :t   True
        :f   False
        :e   empty string
        :n   None
        :z   Ellipsis / NotGiven
        :?   Error, reference to a variable that's not set
             (will parse as NotGiven, for round-trip type safety)

        :xAB Hex integer
        :b01 Binary integer
        :vXY Bytestring, inline
        :yAB Bytestring, hex encoding
        :sAB Bytestring, base64 encoding
        :iXY evaluate XY as a Python expression.
             The 'i' may be missing if XY does not start with a letter.

    Meta elements (delimits elements, SHOULD be in front):

    \b
        :mXX This path is marked with XX (deprecated)
        :R   An alias for the current root
        :Q   An alias for an alternate root
        :P   An alias for another alternate root
        :S   An alias for yet another alternate root

    The empty path is denoted by a single colon. A dotted path that starts
    or ends with a dot, or that contains empty elements (two non-escaped dots,
    one dot followed by a separator) is illegal.

    The alternate slash-path representation uses slashes as separators.
    **Path marks are ignored** when generating the slashed representation.

    Paths can be concatenated with "+", "/" or "|".
    "% n" removes n items from the end.

    All Path objects are read-only.

    The Root paths are context variables. If they are set and an "incoming"
    path has one of them as prefix, said prefix is replaced with a placeholder
    for this root. It is expanded in "slashed" form but not as native
    representation. That way, specific paths can be encoded in a root-free
    form, thus if you ever rename the root, or move entries from one
    MoaT-Link setup to another, everything still works.
    """

    def __init__(self, *a, mark="", scan=False):
        if mark:
            warnings.warn("Marking a path is deprecated")
        if any(isinstance(x,list) for x in a):
            a=tuple(tuple(x) if isinstance(x,list) else x for x in a)
        if a and scan:
            i = 0
            while i < len(a):
                for proxy in _Roots.values():
                    if not proxy:
                        continue

                    if len(a) >= i + len(proxy) and a[i : i + len(proxy)] == proxy:
                        a = a[:i] + (proxy,) + a[i + len(proxy) :]
                        break
                i += 1

        self._data: tuple = a
        if mark is None:
            raise ValueError("Use an empty mark, not 'None'")
        self._mark = mark

    @classmethod
    def build(cls, data, *, mark=""):
        """Optimized shortcut to generate a path from an existing tuple"""
        if mark:
            warnings.warn("Marking a path is deprecated")
        if isinstance(data, Path):
            return data
        if not isinstance(data, tuple) or any(isinstance(x,list) for x in data):
            return cls(*data)
        p = object.__new__(cls)
        p._data = tuple(data)  # noqa:SLF001
        if mark is None:
            raise ValueError("Use an empty mark, not 'None'")
        p._mark = mark  # noqa:SLF001
        return p

    def as_tuple(self):
        """deprecated"""
        return self._data

    @property
    def raw(self):
        """as tuple"""
        return self._data

    @property
    def mark(self):
        "accessor for the path's mark"
        return self._mark

    def startswith(self,path:Path|tuple|list):
        """
        Prefix test
        """
        if isinstance(path,Path):
            path=path._data
        if not isinstance(path,tuple):
            path=tuple(path)

        return self._data[:len(path)] == path

    def with_mark(self, mark=""):
        """Returns the same path with a different mark"""
        if mark:
            warnings.warn("Marking a path is deprecated")
        if mark is None:
            raise ValueError("Use an empty mark, not 'None'")
        return type(self).build(self._data, mark=mark)

    def __str__(self, slash:bool|Literal[2]=False):
        """
        Stringify the path to a dotstring.

        If not slashed: Spaces are escaped somewhat aggressively, for
        better doubleclickability. Do not depend on this.

        If slashed, space escaping is restricted to bytestrings.
        If slash==2, also escape # and +.

        Slash encoding does not work with empty paths; marks are ignored.
        """

        def _escol(x, spaces=True):
            x = x.replace(":", "::")  # must be first
            if slash:
                x = x.replace("/", ":|")
            else:
                x = x.replace(".", ":.")
            if slash == 2:
                x = x.replace("#", ":h")
                x = x.replace("+", ":p")
            if spaces:
                x = x.replace(" ", ":_")
            return x

        res = []
        if self.mark and not slash:
            res.append(":m" + self.mark)
        if self._data is None:
            return ":?"
        if not self._data:
            if slash:
                raise ValueError("Empty paths cannot be slash-coded")
            res.append(":")
        for x in self._data:
            if slash and res:
                res.append("/")

            if isinstance(x, str):
                if slash:
                    res.append(_escol(x, False))
                elif x == "":
                    res.append(":e")
                else:
                    if res:
                        res.append(".")
                    res.append(_escol(x))
            elif x is NotGiven:
                res.append(":z")
            elif x is True:
                res.append(":t")
            elif x is False:
                res.append(":f")
            elif x is None:
                res.append(":n")
            elif isinstance(x, RootPath):
                if not slash:
                    res.append(f":{x.key}")
                else:
                    if not x or len(x) == 0:
                        raise RuntimeError(f"You need to set {x.name}")
                    res.append(x.slashed)

            elif isinstance(x, (bytes, bytearray, memoryview)):
                if all(32 <= b < 127 for b in x):
                    res.append(":v" + _escol(x.decode("ascii"), True))
                else:
                    res.append(":s" + b64encode(x).decode("ascii"))
                    # no hex
            elif isinstance(x, (Path, tuple)):
                if len(x):
                    x = ",".join(repr(y) for y in x)  # noqa: PLW2901
                    res.append(":" + _escol(x) + ("," if len(x)==1 else ""))
                else:
                    x = "()"  # noqa: PLW2901
            else:
                x = repr(x)  # noqa: PLW2901
                if x[0].isalpha():
                    x = "i" + x  # noqa: PLW2901
                res.append(":" + _escol(x))
        return "".join(res)

    @property
    def slashed(self):
        """
        Stringify the path to a slashed string.

        Spaces are not escaped, except in bytestrings.
        """

        return self.__str__(slash=True)

    @property
    def slashed2(self):
        """
        Stringify the path to a slashed string.

        Spaces are not escaped, except in bytestrings.

        This also escapes + and # characters, for use in MQTT publish
        (but possibly-maybe NOT subscribe) paths.
        """

        return self.__str__(slash=2)

    def __getitem__(self, x):
        if isinstance(x, slice) and x.start in (0, None) and x.step in (1, None):
            return type(self)(*self._data[x])
        else:
            return self._data[x]

    def __len__(self):
        return len(self._data)

    def __bool__(self):
        return True

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, Path):
            if self.mark != other.mark:
                return False
            other = other._data
        else:
            try:
                other = tuple(other)
            except TypeError:
                return NotImplemented

        return self._data == other

    def __lt__(self, other):
        other = other._data if isinstance(other, Path) else tuple(other)
        return self._data < other

    def __hash__(self):
        return hash(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __contains__(self, x):
        return x in self._data

    def __mod__(self, other):
        if len(self._data) < other:
            raise ValueError("Path too short")
        return Path(*self._data[:-other], mark=self.mark)

    def _tag_add(self, other):
        if not isinstance(other, Path):
            return self.mark
        if not other.mark:
            return self.mark
        if not self.mark:
            return other.mark
        if self.mark != other.mark:
            raise RuntimeError(
                f"Can't concat paths with different tags: {self.mark} and {other.mark}",
            )
        return self.mark

    def __add__(self, other):
        mark = self._tag_add(other)
        if isinstance(other, Path):
            other = other._data
        elif not isinstance(other, (list, tuple)):
            other = (other,)
        if len(other) == 0:
            if self.mark != mark:
                return self.build(self._data, mark=mark)
            return self
        if isinstance(other[0],Path):
            return type(self)(*self._data, *other[0], *other[1:], mark=mark)
        return type(self)(*self._data, *other, mark=mark)

    def __or__(self, other):
        return self + other

    def __div__(self, other):
        return self + other

    #   def __iadd__(self, other):
    #       mark = self._tag_add(other)
    #       if isinstance(other, Path):
    #           other = other._data
    #       if len(other) > 0:
    #           self._mark = mark
    #           self._data.extend(other)
    #       return self

    def __truediv__(self, other):
        if isinstance(other, Path):
            raise TypeError("You want + not /")
        return Path(*self._data, other, mark=self.mark)

    #   def __itruediv__(self, other):
    #       if isinstance(other, Path):
    #           raise TypeError("You want + not /")
    #       self._data.append(other)

    # TODO add alternate output with hex integers

    def __repr__(self):
        return f"P({str(self)!r})"

    @classmethod
    def from_str(cls, path, *, mark="", scan=False):
        """
        Constructor to build a Path from its string representation.
        """
        res = []
        part: None | bool | str = False
        # non-empty string: accept colon-eval or dot (inline)
        # True: require dot or colon-eval (after :t)
        # False: accept only colon-eval (start)
        # None: accept neither (after dot)

        esc: bool = False
        # marks that an escape char has been seen

        eval_: bool | int = False
        # marks whether the current input shall be evaluated;
        # 2=it's a hex number

        pos = 0
        if isinstance(path, (tuple, list)):
            return cls.build(path, mark=mark)
        if path == ":":
            return cls(mark=mark)

        mp = _RTagRE.match(path)
        if mp:
            if not mark:
                mark = mp[0][2:-1]
            elif mark != mp[0][2:-1]:
                raise SyntaxError(f"Conflicting tags: {mark} vs. {mp[0][2:-1]}")
            return cls(mark=mark)

        def add(x):
            nonlocal part
            if not isinstance(part, str):
                part = ""
            try:
                part += x
            except TypeError:
                raise SyntaxError(f"Cannot add {x!r} at {pos}") from None

        def done(new_part):
            nonlocal part
            nonlocal eval_
            if isinstance(part, str):
                if eval_:
                    try:
                        if eval_ == -1:
                            part = bytes.fromhex(part)
                        elif eval_ == -2:
                            part = part.encode("ascii")
                        elif eval_ == -3:
                            part = b64decode(part.encode("ascii"))
                        elif eval_ > 1:
                            part = int(part, eval_)
                        else:
                            part = path_eval(part)
                    except Exception as exc:
                        raise SyntaxError(f"Cannot eval {part!r} at {pos}") from exc
                    eval_ = False
                res.append(part)
            part = new_part

        def new(x, new_part):
            nonlocal part
            if part is None:
                raise SyntaxError(f"Cannot use {part!r} at {pos}")
            done(new_part)
            res.append(x)

        if path == "":
            raise SyntaxError("The empty string is not a path")
        for e in _PartRE.findall(path):
            if esc:
                esc = False
                if e in ":.":
                    add(e)
                elif e in "z?":
                    new(NotGiven, True)
                elif e == "e":
                    new("", True)
                elif e == "t":
                    new(True, True)
                elif e == "f":
                    new(False, True)
                elif e[0] == "m" and len(e) > 1:
                    done(None)
                    if not mark:
                        mark = e[1:]
                    elif mark != e[1:]:
                        raise SyntaxError(f"Conflicting tags: {mark} vs. {e[1:]} at {pos}")
                    part = True
                elif e == "n":
                    new(None, True)
                elif e in _Roots:
                    new(_Roots[e], True)
                elif e == "_":
                    add(" ")
                elif e[0] == "i":
                    done(None)
                    part = e[1:]
                    eval_ = 1
                elif e[0] == "b":
                    done(None)
                    part = e[1:]
                    eval_ = 2
                elif e[0] == "x":
                    done(None)
                    part = e[1:]
                    eval_ = 16
                elif e[0] == "y":
                    done(None)
                    part = e[1:]
                    eval_ = -1
                elif e[0] == "v":
                    done(None)
                    part = e[1:]
                    eval_ = -2
                elif e[0] == "s":
                    done(None)
                    part = e[1:]
                    eval_ = -3
                else:
                    if part is None:
                        raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                    done("")
                    add(e)
                    eval_ = True
            else:
                if e == ".":
                    if part is None or part is False:
                        raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                    done(None)
                    pos += 1
                    continue
                elif e == ":":
                    esc = True
                    pos += 1
                    continue
                elif part is True:
                    raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                else:
                    add(e)
            pos += len(e)
        if esc or part is None:
            raise SyntaxError(f"Cannot parse {path!r} at {pos}")
        done(None)
        return cls(*res, mark=mark, scan=scan)

    @classmethod
    def from_slashed(cls, path, *, mark=None, scan=True):
        """
        Constructor to build a Path from its slashed string representation.
        """

        res = []

        if isinstance(path, (tuple, list)):
            return cls.build(path, mark=mark)

        def _decol(s):
            s = s.replace(":|", "/").replace(":_", " ")
            s = s.replace(":h", "#").replace(":p", "+")
            return s.replace("::", ":")  # must be last

        marks = 0

        try:
            for pos, p in enumerate(path.split("/")):
                if p == "":
                    res.append("")
                elif p[0] != ":":
                    res.append(_decol(p))
                elif p == ":":
                    pass

                elif p[1] in (":","h","p"):
                    res.append(_decol(p))
                elif p[1] == "b":
                    res.append(int(p[2:], 2))
                elif p[1] == "e":
                    pass
                elif p[1] == "f":
                    if len(p) == 2:
                        res.append(False)
                elif p[1] == "i":
                    res.append(int(p[2:]))
                elif p[1] == "m":
                    if mark is None or mark == p[2:]:
                        mark = p[2:]
                        marks += 1
                elif p[1] == "n":
                    if len(p) == 2:
                        res.append(None)
                elif p[1] == "s":
                    res.append(b64decode(_decol(p[2:]).encode("ascii")))
                elif p[1] == "t":
                    if len(p) == 2:
                        res.append(True)
                elif p[1] == "v":
                    res.append(_decol(p[2:]).encode("ascii"))
                elif p[1] == "x":
                    res.append(int(p[2:], 16))
                elif p[1] == "y":
                    res.append(bytes.fromhex(p[2:]))

                else:
                    res.append(path_eval(p[1:]))

                if len(res) != pos + 1 - marks:
                    raise RuntimeError("Slashed-Path syntax")

        except Exception as exc:
            raise SyntaxError(f"Cannot eval {path!r}, part {pos + 1}") from exc

        if mark is None:
            mark = ""
        r = cls(*res, mark=mark, scan=scan)
        return r

    @classmethod
    def _make(cls, loader, node):
        value = loader.construct_scalar(node)
        return cls.from_str(value)


    def apply(self, path:Path) -> Path:
        """
        Construct a new path that replaces pattern tuples in @path with
        the referred-to entries in @self.

        @path is returned unchanged if @self doesn't contain any patterns.

        Thus:

        >>> p = Path("w.x.y.z")
        >>> P("a:2,.b").apply(p)
        P("a.x.b")
        >>> P("a:3,4.b").apply(p)
        P("a.y.z.b")

        Elements are numbered starting from 1 (left) or -1 (right).
        """
        if not any(isinstance(x,tuple) for x in self._data):
            return self

        # We might want to cache this …
        res = []
        for p in self._data:
            if not isinstance(p,tuple) or len(p) not in (1,2,3):
                res.append(p)
                continue

            p = list(p)
            if p[0] > 0:
                p[0] -= 1
            if len(p) == 1:
                res.append(path[p[0]])
            else:
                if p[1] < 0:
                    p[1] += 1
                    if p[1] == 0:  # was: -1
                        p[1] = None
                res.extend(path[slice(*p)])
        return Path.build(res)


class P(Path):
    """
    A Path subclass that delegates to `Path.from_str`.

    For idempotency (required by ``click``) it transparently accepts `Path`
    objects.

    Scanning for prefixes is disabled. Use this class for paths embedded in
    MoaT code.
    """

    def __new__(cls, path, *, mark="", scan=False):  # noqa:D102
        if isinstance(path, Path):
            if path.mark != mark:
                path = Path(*path, mark=mark, scan=scan)
            return path
        return Path.from_str(path, mark=mark, scan=scan)


class PP(Path):
    """
    A Path subclass that delegates to `Path.from_str`.

    This is identical to `P` except that scanning for prefixes is enabled.
    Use this class for command-line processing.
    """

    def __new__(cls, path, *, mark="", scan=True):
        if isinstance(path, Path):
            if path.mark != mark:
                path = Path(*path, mark=mark, scan=scan)
            return path
        return Path.from_str(path, mark=mark, scan=scan)


class PS(Path):
    """
    A Path subclass that delegates to `Path.from_path`.

    For idempotency (required by ``click``) it transparently accepts `Path`
    objects.
    """

    def __new__(cls, path, *, mark=""):  # noqa:D102
        if isinstance(path, Path):
            if path.mark != mark:
                path = Path(*path, mark=mark)
            return path
        return Path.from_slashed(path, mark=mark, scan=True)


def logger_for(path: Path):
    """
    Create a logger for this ``path``.

    The logger always starts with your main module name, with special
    treatment for paths starting with null or your-module-with-leading-dot.
    Thus if you import this code as "foo.util"::

        this path returns a logger for
        ========= ====================
        :         foo.root
        :n        foo.meta
        :n.x.y    foo.meta.x.y
        :.foo     foo.sub
        :.foo.z   foo.sub.z
        foo       foo
        foo.a.b   foo.a.b
        foo.:.a.b foo.a.b
        bar       foo.at.bar
        bar.c.d   foo.at.bar.c.d

    All elements in the path should be strings with no leading or trailing
    dot, though the first element may start with a dot or be None.

    """
    this = __name__.split(".", 1)[0]
    if len(path) == 0:
        p = f"{this}.root"
    elif path[0] is None:
        p = f"{this}.meta"
    elif path[0] == f".{this}":
        p = f"{this}.sub"
    elif path[0] == this:
        p = this
    else:
        p = f"{this}.at.{path[0]}"
    if len(path) > 1:
        p += "." + ".".join(str(x) or "-" for x in path[1:])
    p = p.replace("..", ".")
    return logging.getLogger(p)


class PathShortener:
    """This class shortens path entries so that the initial components that
    are equal to the last-used path (or the original base) are skipped.

    It is illegal to path-shorten messages whose path does not start with
    the initial prefix.

    Example: The sequence

        a b
        a b c d
        a b c e f
        a b c e g h
        a b c i
        a j k

    is shortened to

        0 a b
        2 c d
        3 e f
        4 g h
        3 i
        1 j k

    Usage::

        >>> d = PathShortener()
        >>> d.short(P('a.b.c.d'))
        (4, ('a','b',c','d'))
        >>> d.short(P('a.b.c.e.f'))
        (3, ('e','f'))
        {'depth':1, 'path':['e','f']}

    Alternate, somewhat-deprecated usage::
        >>> d = PathShortener(['a','b'])
        >>> d({'path': 'a b c d'.split})
        {'depth':0, 'path':['c','d']}
        >>> d({'path': 'a b c e f'.split})
        {'depth':1, 'path':['e','f']}

    Note that the input dict in the second example is modified in-place.

    Using a prefix is deprecated.

    Caution: this shortener ignores path marks.
    """

    def __init__(self, prefix:Path|list|tuple=Path()):  # noqa:B008
        self.prefix = prefix if isinstance(prefix,Path) else Path.build(prefix)
        self.depth = len(prefix)
        self.path = []

    def short(self, p: Path) -> tuple[int, Path]:
        """shortens the given path"""
        if self.depth and list(p[: self.depth]) != list(self.prefix):
            raise RuntimeError(f"Wrong prefix: has {p!r}, want {self.prefix!r}")

        p = p[self.depth :]
        cdepth = min(len(p), len(self.path))
        for i in range(cdepth):
            if p[i] != self.path[i]:
                cdepth = i
                break
        self.path = p
        return cdepth, p[cdepth:]

    def __call__(self, res: dict):
        "shortens the 'path' element in @res"
        try:
            p = res["path"]
        except KeyError:
            return
        cdepth, p = self.short(p)
        res["path"] = p
        res["depth"] = cdepth
        return res


class PathLongener:
    """
    This reverts the operation of a PathShortener. You need to pass the
    same prefix in.

    Calling a PathLongener with a dict without ``depth`` or ``path``
    attributes is a no-op.

    Using a prefix is deprecated.

    Caution: this longener ignores path marks.
    """

    cls = Path

    def __init__(self, prefix: Path | tuple | list = ()):
        if isinstance(prefix, Path):
            self.cls = type(prefix)
            prefix = prefix.raw
        elif isinstance(prefix,list):
            prefix=tuple(prefix)
        self.depth = len(prefix)
        self.path = prefix

    def long(self, d: int | None, p: Path):
        """Expand a given path suffix"""
        p = tuple(p)
        if d is None:
            return p
        p = self.cls.build(self.path[: self.depth + d] + p)
        self.path = p
        return p

    def __call__(self, res):
        "expands the 'path' element in @res"
        p = res.get("path", None)
        if p is None:
            return
        d = res.pop("depth", None)
        if d is None:
            return
        res["path"] = self.long(d, p)
        return res


# path_eval is a simple "eval" replacement to implement resolving
# expressions in paths. While it can be used for math, its primary function
# is to process tuples.
_eval = simpleeval.SimpleEval(functions={})
_eval.nodes[ast.Tuple] = lambda node: tuple(
    _eval._eval(x)  # noqa:SLF001
    for x in node.elts
)
path_eval = _eval.eval


# Here we declare our bunch of "root" variables.

Root = ContextVar[Path|None]("Root", default=None)


class RootPath(Path):
    """
    Wraps access to a contextvar that points to a Path.

    The problem is that the contextvar's ID is not stable. However proxying
    it must be consistent regardless of its content.
    """

    _mark = ""

    def __init__(self, key, var, name):
        self._key = key
        self._var = var
        self._name = name

    @property
    def name(self):
        "name"
        return self._name

    @property
    def key(self):
        "name of the contextvar"
        return self._key

    def __bool__(self):
        "check if the contextvar is set"
        p = self._var.get()
        return p is not None

    @property
    def _data(self):
        p = self._var.get()
        if p is None:
            return None
        return self._var.get()._data  # noqa:SLF001


_root = RootPath("R", Root, "Root")
as_proxy("R", _root)
_Roots = {"R": _root}

for _idx in "SPQ":  # and R. Yes I know.
    _name = f"{_idx}_Root"
    _ctx = ContextVar(_name, default=None)
    _path = RootPath(_idx, _ctx, _name)
    _ctx.set(Path("XXX",_idx,"XXX"))

    globals()[_name] = _ctx
    __all__ += [_name]  # noqa:PLE0604

    _Roots[_idx] = _path
    as_proxy(f"_P{_idx}", _path)

del _idx, _name, _ctx, _path



"""
This module contains msgpack-based helper functions and classes.

TODO add CBOR.
"""

from __future__ import annotations

import anyio

from moat.lib.codec import Codec
from pathlib import Path as FSPath

__all__ = ["MsgReader", "MsgWriter"]


class _MsgRW:
    """
    Common base class for :class:`MsgReader` and :class:`MsgWriter`.
    """

    _mode: str = None

    def __init__(self, path:anyio.Path|FSPath|str|None=None, stream=None, codec:Codec|str=None):
        if (path is None) == (stream is None):
            raise RuntimeError("You need to specify either path or stream")

        if codec is None:
            raise ValueError("No default codec")
        if not isinstance(codec, Codec):
            from moat.util import get_codec
            codec = get_codec(codec)

        if isinstance(path,anyio.Path):
            pass
        elif path is not None:
            path = anyio.Path(path)
        self.path = path

        self.stream = stream
        self.codec = codec

    async def __aenter__(self):
        if self.path is not None:
            p = self.path
            if p == "-":
                if self._mode[0] == "r":  # pylint: disable=unsubscriptable-object
                    p = "/dev/stdin"
                else:
                    p = "/dev/stdout"
            self.stream = await anyio.open_file(p, self._mode)
        return self

    async def __aexit__(self, *tb):
        if self.path is not None:
            with anyio.CancelScope(shield=True):
                await self.stream.aclose()


class MsgReader(_MsgRW):
    """Read a stream of messages (encoded with some codec) from a file.

    Usage::

        async with MsgReader(path="/tmp/msgs.pack") as f:
            async for msg in f:
                process(msg)

    Arguments:
      buflen (int): The read buffer size. Defaults to 4k.
      path (str): the file to write to.
      stream: the stream to write to.

    Exactly one of ``path`` and ``stream`` must be used.
    """

    _mode = "rb"

    def __init__(self, *a, buflen=4096, **kw):
        super().__init__(*a, **kw)
        self.buflen = buflen

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                return next(self.codec)
            except StopIteration:
                pass

            d = await self.stream.read(self.buflen)
            if d == b"":
                raise StopAsyncIteration
            self.codec.feed(d)


class MsgWriter(_MsgRW):
    """Write a stream of messages to a file (encoded with some codec).

    Usage::

        async with MsgWriter("/tmp/msgs.pack", codec="std-cbor") as f:
            for msg in some_source_of_messages():  # or "async for"
                await f(msg)

    Arguments:
      buflen (int): The buffer size. Defaults to 64k.
      path (str): the file to write to.
      stream: the stream to write to.

    Exactly one of ``path`` and ``stream`` must be used.

    The stream is buffered. Call :meth:`flush` to flush the buffer.
    """

    _mode = "wb"

    def __init__(self, *a, buflen=65536, **kw):
        super().__init__(*a, **kw)

        self.buf = []
        self.buflen = buflen
        self.curlen = 0
        self.excess = 0

    async def __aexit__(self, *tb):
        with anyio.fail_after(2, shield=True):
            if self.buf:
                await self.stream.write(b"".join(self.buf))
            await super().__aexit__(*tb)

    async def __call__(self, msg):
        """Write a message (bytes) to the buffer.

        Flushing writes a multiple of ``buflen`` bytes."""
        msg = self.codec.encode(msg)
        self.buf.append(msg)
        self.curlen += len(msg)
        if self.curlen + self.excess >= self.buflen:
            buf = b"".join(self.buf)
            pos = self.buflen * ((self.curlen + self.excess) // self.buflen) - self.excess
            assert pos > 0
            wb, buf = buf[:pos], buf[pos:]
            self.curlen = len(buf)
            self.buf = [buf]
            self.excess = 0
            await self.stream.write(wb)

    async def flush(self, force=True):
        """Flush the buffer.

        @force: do write partial data.
        """
        if self.buf:
            buf = b"".join(self.buf)
            self.buf = []
            self.excess = (self.excess + len(buf)) % self.buflen
            await self.stream.write(buf)
            if force:
                await self.stream.flush()



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.

Extension types defined here:
2: contains raw bytes, interpreted as unsigned bignum
3: Path, as a msgpack object stream of its elements
4: contains raw bytes, interpreted as UTF-8, returned as (named) Proxy object
5: object constructor
6: marked Path
"""

from __future__ import annotations


from moat.lib.codec.msgpack import Codec

from ._msgpack import std_ext, StdMsgpack

__all__ = ["std_ext", "StdMsgpack"]


Codec = StdMsgpack


@std_ext.encoder(2, int)
def _enc_int(codec, n):
    codec  # noqa:B018
    return n.to_bytes((n.bit_length() + 7) // 8, "big")


@std_ext.decoder(2)
def _dec_bignum(codec, data):
    codec  # noqa:B018
    return int.from_bytes(data, "big")



"""
Broadcasting support
"""

from __future__ import annotations

from weakref import WeakSet

from attrs import define, field

from .compat import EndOfStream, WouldBlock
from . import NotGiven
from .queue import Queue

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Literal, Self

# TODO build something nicer
try:
    import logging
except ImportError:
    logger = None
else:
    logger = logging.getLogger(__name__)

__all__ = [
    "Broadcaster",
    "BroadcastReader",
    "LostData",
]


class LostData(Exception):
    """
    Indicator of data loss.

    Attribute ``n`` contains the number of dropped messages.
    """

    def __init__(self, n):
        self.n = n


@define
class BroadcastReader: ## TYPE [TData]:
    """
    The read side of a broadcaster.

    Simply iterate over it.

    Warning: The iterator may raise ``LostData`` exceptions.
    These contain the number of messages that have been dropped
    due to the reader being too slow.

    Readers may be called to inject values.
    """

    parent: Broadcaster = field()
    length: int = field(default=1)
    value: TData | Literal[NotGiven] = field(default=NotGiven, init=False)

    loss: int = field(init=False, default=0)
    _q: Queue = field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.length <= 0:
            raise RuntimeError("Length must be at least one")
        self._q = Queue(self.length)

    def __hash__(self):
        return id(self)

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> TData:
        if self.loss > 0:
            n, self.loss = self.loss, 0
            raise LostData(n)

        try:
            return await self._q.get()
        except (AttributeError, EndOfStream, EOFError):
            raise StopAsyncIteration from None

    def flush(self) -> None:
        """
        Clean the queue.

        Useful for re-sync after you get a `LostData` error.
        """
        try:
            while True:
                self._q.get_nowait()
        except WouldBlock:
            return

    def __call__(self, value: TData) -> None:
        """enqueue a value, to this reader only"""
        try:
            self._q.put_nowait(value)
        except WouldBlock:
            x = self._q.get_nowait()
            if logger is not None:
                logger.debug("Dropped: %r", x)
            self._q.put_nowait(value)
            self.loss += 1

    def _close(self) -> None:
        self._q.close_writer()

    def close(self) -> None:
        "close this reader, detaching it from its parent"
        self._close()
        self.parent._closed_reader(self)  # noqa:SLF001 pylint: disable=protected-access

    async def aclose(self) -> None:
        "close this reader, detaching it from its parent"
        self.close()


@define
class Broadcaster: ## TYPE [TData]:
    """
    A simple broadcaster. Messages will be sent to all readers.

    @length is each reader's default queue length.

    If @send_last is set, a new reader immediately gets the last-sent
    value. Otherwise it waits for new data.

    If a queue is full, the oldest message will be discarded. Readers will
    then get a LostData exception that contains the number of dropped
    messages.

    To write, open a context manager (sync or async) and call with a value.

    To read, async-iterate::

        async def rdr(bcr: BroadcastReader|Broadcaster):
            async for msg in bcr:
                print(msg)

        async with anyio.create_task_group() as tg, Broadcaster() as bc:
            tg.start_soon(rdr, aiter(bc))
            for x in range(5):
                bc(x)
                await anyio.sleep(0.01)
            bc(42)

    To safely re-sync, do something like this::

        bcr.flush()
        x = fetch_consistent_state()  # this really should be sync
        bcr(x)  # if this fails, you've got a problem
        while True:
            y = await anext(bcr)
            if x is y:
                break

    """

    length: int = field(default=1)
    send_last: bool = field(default=False)

    _rdr: WeakSet[BroadcastReader] | None = field(init=False, default=None, repr=False)
    value: TData | Literal[NotGiven] = field(init=False, default=NotGiven)

    def open(self) -> Self:
        """Open the broadcaster.

        Consider using a context instead of this method.
        """
        if self._rdr is not None:
            raise RuntimeError("already entered/opened")
        self._rdr = WeakSet()
        return self

    def __enter__(self) -> Self:
        return self.open()

    async def __aenter__(self) -> Self:
        return self.open()

    def __exit__(self, *tb) -> None:
        self.close()

    async def __aexit__(self, *tb) -> None:
        self.close()

    def _closed_reader(self, reader) -> None:
        assert self._rdr is not None

        self._rdr.remove(reader)

    def __aiter__(self) -> BroadcastReader[TData]:
        """Create a reader with the predefined queue length"""
        assert self._rdr is not None

        r: BroadcastReader[TData] = BroadcastReader(self, self.length)
        self._rdr.add(r)
        if self.send_last and self.value is not NotGiven:
            r(cast("TData", self.value))
        return aiter(r)

    def reader(self, length: int, send_last:bool|None=None) -> BroadcastReader[TData]:
        """Create a reader with an explicit queue length"""
        assert self._rdr is not None

        if send_last is None:
            send_last = self.send_last
        r: BroadcastReader[TData] = BroadcastReader(self, length)
        self._rdr.add(r)
        if send_last and self.value is not NotGiven:
            if not length:
                raise ValueError("This would deadlock. Use length>0.")
            r(cast("TData", self.value))
        return aiter(r)

    def __call__(self, value: TData) -> None:
        """Enqueue a value to all readers"""
        assert self._rdr is not None

        self.value = value
        for r in self._rdr:
            r(value)

    async def read(self) -> TData:
        "gets the last value (waits until there is one)"
        if self.value is NotGiven:
            return await anext(aiter(self))
        return self.value

    def close(self):
        "Close the broadcaster. No more writing."
        if self._rdr is not None:
            for r in self._rdr:
                r._close()  # noqa:SLF001 pylint: disable=protected-access
            self._rdr = None



"""
Implements an alert object and dispatcher.

Alerts are exceptions that can be iterated or waited on.

Waiting will succeed when the alarm is resolved/closed.
Iterating will receive new state and end when the alert is resolved.

The dispatcher is a context manager that can be iterated to receive new
alerts.
"""

from __future__ import annotations

from contextlib import asynccontextmanager

from .broadcast import Broadcaster, BroadcastReader
from .compat import Event, TaskGroup
from .ctx import CtxObj

__all__ = [
    "Alert",
    "BaseAlert",
    "RepeatAlert",
    "AlertHandler",
    "AlertMixin",
    "AlertCollector",
]


class BaseAlert(Exception):
    """Alert, initial OR repeat wrapper"""

    # pylint:disable=unnecessary-pass


class Alert(BaseAlert):
    """
    This is an iteratable alert: it can be iterated for new state, or
    closed when complete.

    The iterator terminates when the condition has passed.

    Intended use: subclass this, distribute using an AlarmHandler mix-in.
    """

    state: list = None
    evt: Event = None

    def __init__(self, *data):
        super().__init__(*data)
        self.evt = Event()
        self.q = Broadcaster()
        self.q.__enter__()
        self.set(data)

    def __aiter__(self):
        return self.q.__aiter__()

    def set(self, *data) -> None:
        """
        Set or update the state of this alert.
        """
        self.q(data)

    def resolve(self) -> None:
        """
        End this alert.
        """
        self.q.close()
        self.evt.set()

    def __await__(self):
        return self.wait().__await__()

    async def wait(self) -> None:
        """
        Wait for the alert to end.
        """
        await self.evt.wait()

    def __del__(self):
        self.q.__exit__(None, None, None)


class RepeatAlert(BaseAlert):
    """
    If an existing alert is re-raised, it is wrapped in a RepeatAlert so as
    to not break the existing alert's traceback and related data.
    """

    def __init__(self, err):
        self.error = err


class AlertHandler:
    """
    Collect open alerts.

    This helper class keeps track of open alerts and lets multiple clients
    receive them.
    """

    def __init__(self):
        self.__alarms = {}
        self.__q = Broadcaster()

    def __enter__(self):
        self.__q.__enter__()
        return self

    async def __aenter__(self):
        self.__q.__enter__()
        return self

    async def __aexit__(self, *tb):
        self.__q.__exit__(*tb)

    def __exit__(self, *tb):
        self.__q.__exit__(*tb)

    def reader(self, length=3):
        """
        Returns an async iterator for new alerts.
        """
        return self.__q.reader(length=length)

    def alerts(self):
        """List open alerts"""
        return self.__alarms.values()

    def alert_(self, cls, *msg) -> Alert:
        """
        Alarm trigger.

        Alarms are static: you can update an alarm with new data.

        To clear an alarm, pass in no data.
        """
        if cls in self.__alarms:
            if msg:
                err = self.__alarms[cls]
                err(*msg)
            else:
                err = self.__alarms.pop(cls)
                err.resolve()
            return err
        elif not msg:
            return None
        err = cls(*msg)
        self.__alarms[cls] = err
        self.__q(err)
        return err

    def raise_(self, cls, *msg):
        """
        Raises this alert as an exception, in addition to queuing it.

        An updated exception will be wrapped in a RepeatAlert.
        """
        # This is a mangled copy of `alert_`.
        if cls in self.__alarms:
            if msg:
                err = self.__alarms[cls]
                err(*msg)
                raise RepeatAlert(err)
            else:
                err = self.__alarms.pop(cls)
                err.close()
                return
        elif not msg:
            return
        err = cls(*msg)
        self.__alarms[cls] = err
        self.__q(err)
        raise err

    def watch_(self, length=3) -> BroadcastReader:
        """
        Monitor for new alerts.
        """
        return self.__q.reader(length)


class AlertMixin:
    """
    Use this mixin to equip your class with alert handling.

    The `AlertHandler` is stored in the ``_alerts`` attribute.
    """

    def __init__(self, *a, **kw):
        self._alerts = AlertHandler()
        super().__init__(*a, **kw)

    def alerts(self, length=3):
        """
        Return an async iterator for new alerts.

        Pass @length to change the maximal queue length (default 3)
        """
        return self._alerts.reader(length)

    def current_alerts(self):
        """
        Return a list of currently-active alerts.

        The result is an iterator. Don't switch tasks while it is active.
        """
        return self._alerts.alerts()

    def set_alert(self, cls, *msg):
        """
        Alert trigger/updater.

        To clear an alert, pass in no data.
        """
        self._alerts.alert_(cls, *msg)

    def raise_alert(self, cls, *msg):
        """
        Raising alert trigger/updater.

        To clear an alert, pass in no data.
        """
        self._alerts.raise_(cls, *msg)


class AlertCollector(CtxObj):
    """
    This context manager stores a variety of alerts, or in fact any other
    object that can be ``await``ed. It is basically the read-side
    counterpart of an `AlertHandler`.

    Awaiting it will delay until all alerts that have been fed to it are
    resolved.

    If the object itself needs to be awaited, or a function other than
    ``wait``, set @access to an appropriate Lambda:

        evtents = AlertController(access=lambda x:x)
    """

    _tg = None
    _working = None

    def __init__(self, access=lambda x: x.wait()):
        self.objs = set()
        self.access = access

        self.non_empty = Event()
        self.evt = Event()
        self.evt.set()

    # if non_empty is not None:
    #    .evt is set
    #    the system is idle.
    # otherwise
    #    .evt is not set
    #    .objs and _working contain the alerts to be waited on.
    #
    # The transition from empty to non-empty is the job of ``.add()``
    # (which doesn't depend on the context to be active).

    def __repr__(self):
        if self.objs or self._working:
            evts = " ".join(self.objs) + " " + str(self._working)
            return f"<{self.__class__.__name__}: {evts}>"
        else:
            return f"<{self.__class__.__name__}: empty>"

    async def wait(self):
        "Wait until all alerts have ben resolved."
        await self.evt.wait()

    def __await__(self):
        return self.evt.wait().__await__()

    async def _runner(self):
        while True:
            if self.objs:
                self._working = w = self.objs.pop()
                try:
                    await self.access(w)
                except BaseException:
                    self.objs.add(w)
                    raise
                finally:
                    self._working = None
            else:
                self.evt.set()
                if self.non_empty is None:
                    self.non_empty = Event()
                await self.non_empty.wait()
                self.non_empty = None

    async def wait_busy(self):
        """
        wait until an alert has been raised.
        """
        if self.non_empty is None:
            return
        await self.non_empty.wait()

    def is_busy(self):
        """
        Check if this collector contains unresolved alerts.

        May be spuriously positive if the internal task is slow, or not
        running.
        """
        return self.non_empty is None

    @asynccontextmanager
    async def _ctx(self):
        async with TaskGroup() as self._tg:
            self._tg.start_soon(self._runner)
            yield self
            self._tg.cancel_scope.cancel()

    def add(self, thing):
        """Wait for this event."""
        if self.non_empty is not None:
            self.evt = Event()
            self.non_empty.set()
        self.objs.add(thing)



"Support for merge-in-place of dict contents"

from __future__ import annotations

__all__ = ["merge"]


from . import NotGiven


def _merge_dict(d, other, drop=False, replace=True):
    for key, value in other.items():
        if value is NotGiven:
            d.pop(key, None)
        elif key in d:
            d[key] = _merge_one(d[key], value, drop=drop, replace=replace)
        else:
            d[key] = value

    if drop:
        keys = []
        for k in d.keys():  # noqa:SIM118
            if k not in other:
                keys.append(k)
        for k in keys:
            del d[k]


def _merge_list(item, value, drop=False, replace=True):
    off = 0
    if isinstance(value, (list, tuple)):
        # two lists
        lim = len(item)
        for i in range(min(lim, len(value))):
            if value[i] is NotGiven:
                item.pop(i - off)
                off += 1
            else:
                item[i - off] = _merge_one(item[i - off], value[i], drop=drop, replace=replace)

        while len(item) + off < len(value):
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)

        if drop:
            while len(item) + off > len(value):
                item.pop()

    else:
        # a list, and a dict with updated/new/deleted values
        for i in range(len(item)):
            if i in value:
                val = value[i]
                if val is NotGiven:
                    item.pop(i - off)
                    off += 1
                else:
                    item[i - off] = _merge_one(item[i - off], val, drop=drop, replace=replace)

        while len(item) + off in value:
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)


def _merge_one(d, other, drop=False, replace=True):
    if isinstance(d, dict):
        if isinstance(other, dict):
            _merge_dict(d, other, drop=drop, replace=replace)
        else:
            return other if replace else d
    elif isinstance(d, list):
        if isinstance(other, (dict, list, tuple)):
            _merge_list(d, other, drop=drop, replace=replace)
        else:
            return other if replace else d
    else:
        if replace:
            return d if other is None else other
        else:
            return other if d is None else d
    return d


def merge(d, *others, drop=False, replace=True):
    """
    Deep-merge a "source" and one or more "replacement" data structures.

    In contrast to `combine_dict`, the source is modified in-place.

    Rules:
    * two dicts: recurse into same-key items, and adds new items
    * two lists: recurse into same-index items
      append to the source if the replacement is longer
    * a list and a dict treats the dict as a sparse list. The value of numeric keys just
      beyond the list's length are appended, others are ignored.
    * a replacement value of `NotGiven` deletes the source entry
    * otherwise, use the second argument if it is not None, otherwise the first
      * except that if "replace" is False, these are swapped

    If "drop" is given, delete source keys that are not in the destination.
    This is useful for in-place replacements.
    """
    for other in others:
        d = _merge_one(d, other, drop=drop, replace=replace)
    return d



"""
This module contains a heap of somewhat-random helper functions
and classes which are used throughout MoaT (and beyond)
but don't get their own package because they're too small,
or too interrelated … or the author was too lazy.
"""
# TODO split this up

# pylint: disable=cyclic-import,wrong-import-position
from __future__ import annotations

import logging as _logging

_log = _logging.getLogger(__name__)

NotGiven = Ellipsis

from .dict import attrdict  # noqa: E402, F401
from .config import CFG, ensure_cfg  # noqa: E402, F401

from .alert import *  # noqa: F403, E402  # isort:skip
from .impl import *  # noqa: F403, E402  # isort:skip
from .dict import *  # noqa: F403, E402  # isort:skip
from .merge import *  # noqa: F403, E402  # isort:skip
from .misc import *  # noqa: F403, E402  # isort:skip
from .random import *  # noqa: F403, E402  # isort:skip
from .inexact import *  # noqa: F403, E402  # isort:skip

from moat.lib.codec.proxy import as_proxy  # noqa: F403, E402  # isort:skip

as_proxy("_", NotGiven)


try:
    from .event import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .event)", exc)

try:
    from .ctx import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .ctx)", exc)

try:
    from .queue import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .queue)", exc)

try:
    from .module import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .module)", exc)

try:
    from .msg import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .msg)", exc)

try:
    from .msgpack import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .msgpack)", exc)

try:
    from .path import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .path)", exc)

try:
    from .server import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .server)", exc)

try:
    from .spawn import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .spawn)", exc)

try:
    from .systemd import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .systemd)", exc)

try:
    from .yaml import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .yaml)", exc)

try:
    from .exc import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s (importing .exc)", exc)

from .main import *  # noqa: F403, E402



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

__all__ = ["val2pos", "pos2val", "srepr", "OutOfData", "_add_obj", "get_codec"]


class OutOfData(EOFError):
    pass

def get_codec(name:str|Codec):
    "Codec loader; replaces 'std-' prefix with 'moat.util.'"
    from moat.lib.codec import get_codec as _get_codec, Codec

    if isinstance(name,Codec):
        return name
    if name is None:
        name = "noop"
    elif name[0:4] == "std-":
        name = "moat.util." + name[4:]
    return _get_codec(name)


def val2pos(a: float, b: float, c: float, /, clamp: bool = False):
    """
    Return the position of @b within [@a…@c].

        b==a => 0
        b==c => 1
        b==(a+c)/2 => 0.5

    If @clamp is set, the return value is limited to [0…1] even if @b is
    outside a…c.

    @a and @c may not be equal.
    """

    res = (a - b) / (a - c)
    if clamp:
        if res < 0:
            res = 0
        elif res > 1:
            res = 1
    return res


def pos2val(a: float, b: float, c: float, /, clamp: bool = False):
    """
    Return the value of @b within [@a…@c].

        a,0,c => a
        a,1,c => c
        a,0.5,c => (a+c)/2
        a,-1,c => 2*a-c
        a,2,c => 2*c-a

    If @clamp is set, the input is limited to [0…1].

    @a and @c may not be equal.
    """

    if clamp:
        if b < 0:
            b = 0
        elif b > 1:
            b = 1

    return a + b * (c - a)


def drepr(k, v):
    k = str(k)
    if v is None:
        return "?" + k
    if v is False:
        return "!" + k
    if v is True:
        return k
    return f"{k}={srepr(v)}"


def srepr(x):
    "short repr of possibly-complex objects"
    if isinstance(x, set):
        if not x:
            return "∅"
        else:
            return "⊕".join(srepr(v) for v in x)
    if isinstance(x, tuple):
        return "(" + ",".join(srepr(v) for v in x) + ")"
    if isinstance(x, list):
        return "(" + ",".join(srepr(v) for v in x) + ")"
    if isinstance(x, dict):
        return "{" + ",".join(drepr(k, v) for k, v in x.items()) + "}"
    try:
        d = vars(x)
    except TypeError:
        return str(x)
    else:
        return f"{type(x).__name__}{srepr(d)}"


def _add_obj(a, b):
    """add attributes of B to A if they're missing"""
    for k in dir(b):
        if not hasattr(a, k):
            setattr(a, k, getattr(b, k))



"""
This module imports possibly-partial configuration snippets.
"""

from __future__ import annotations

from importlib import import_module
from pathlib import Path as FSPath

from .yaml import yload
from .dict import attrdict
from .merge import merge

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .path import Path

__all__ = ["CFG", "ensure_cfg"]
CFG = attrdict()


def ensure_cfg(path: str | Path, cfg=CFG) -> dict:
    """
    Ensure that a submodule's default configuration is available.
    """
    if isinstance(path, str):
        path = path.split(".")

    def _load(cfg, p):
        ext = import_module(p)

        try:
            p = ext.__path__
        except AttributeError:
            p = (str(FSPath(ext.__file__).parent),)

        for d in p:
            fn = FSPath(d) / "_cfg.yaml"
            if fn.is_file():
                merge(cfg, yload(fn, attr=True))

    try:
        EXT = cfg.setdefault("ext", attrdict())
        EXT["moat"] = cfg

        if "logging" not in cfg:
            _load(cfg, "moat")

        cc = EXT
        for n in range(len(path)):
            cc = cc.setdefault(path[n], attrdict())
            if "_def" in cc:
                continue
            _load(cc, ".".join(path[: n + 1]))
            cc["_def"] = True

    finally:
        del EXT["moat"]

    return cfg



# Trio step monitors

from __future__ import annotations
from trio.abc import Instrument
from trio import Cancelled
from trio._core._instrumentation import Instruments
import sys
from outcome import Value, Error


class Instr(Instrument):
    exited = None

    def before_task_step(self, task):
        print("BEF:", task, task._next_send, file=sys.stderr)

    def after_task_step(self, task):
        print("AFT:", task, task._next_send, file=sys.stderr)

    def task_exited(self, task, outcome=None):
        print("EXIT", task, outcome, file=sys.stderr)


def hookup(runner=None):
    if runner is None:
        import inspect as i

        f = i.currentframe().f_back
        while "runner" not in f.f_locals:
            f = f.f_back
            if f is None:
                return  # no Trie
        runner = f.f_locals["runner"]

    i = Instr()
    ig = Instruments([i])
    runner.instruments = ig



from __future__ import annotations


def capture(sync_fn, *args, **kwargs):
    """Run ``sync_fn(*args, **kwargs)`` and capture the result."""
    try:
        return Value(sync_fn(*args, **kwargs))
    except Exception as exc:
        return Error(exc)


async def acapture(async_fn, *args, **kwargs):
    """Run ``await async_fn(*args, **kwargs)`` and capture the result."""
    try:
        return Value(await async_fn(*args, **kwargs))
    except Exception as exc:
        return Error(exc)


class Outcome:
    """An abstract class representing the result of a Python computation."""

    pass


class Value(Outcome):
    """Concrete :class:`Outcome` subclass representing a regular value."""

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"Value({self.value!r})"

    def unwrap(self):
        return self.value


class Error(Outcome):
    """Concrete :class:`Outcome` subclass representing a raised exception."""

    def __init__(self, error):
        self.error = error

    def __repr__(self):
        return f"Error({self.error!r})"

    def unwrap(self):
        captured_error = self.error
        try:
            raise captured_error
        finally:
            del captured_error, self



"""
Floats that compare with a delta.
"""

from __future__ import annotations

from math import isclose

__all__ = ["InexactFloat"]


class InexactFloat(float):
    """Float wrapper used for inexact comparison of float record elements."""

    __slots__ = ("abs", "rel", "digits")

    def __new__(cls, val, rel=1e-06, abs=1e-12, digits=3):
        res = super().__new__(cls, val)
        res.rel = rel
        res.abs = abs
        res.digits = digits
        return res

    def __hash__(self):
        raise TypeError(f"unhashable type: {type(self)}")

    def __eq__(self, b):
        """Check if the two floats are mostly-equal."""
        return isclose(self, b, rel_tol=self.rel, abs_tol=self.abs)

    def __ne__(self, b):
        return not isclose(self, b, rel_tol=self.rel, abs_tol=self.abs)

    def __lt__(self, other):
        return super().__lt__(other) and self != other

    def __gt__(self, other):
        return super().__gt__(other) and self != other

    def __le__(self, other):
        return super().__le__(other) or self == other

    def __ge__(self, other):
        return super().__ge__(other) or self == other

    # --repr
    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __str__(self):
        return f"{round(self, self.digits):.{self.digits}f}"



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import logging
import sys
from collections import deque
from contextlib import nullcontext, suppress
from getpass import getpass
from math import log10

__all__ = [
    "NoneType",
    "singleton",
    "TimeOnlyFormatter",
    "count",
    "acount",
    "Cache",
    "NoLock",
    "OptCtx",
    "digits",
    "num2byte",
    "byte2num",
    "split_arg",
    "num2id",
    "import_",
    "load_from_cfg",
]

NoneType = type(None)

NoLock = nullcontext()


class OptCtx:
    """
    Optional context. Unlike `contextlib.nullcontext` this doesn't return a
    fixed value; instead it delegates to the wrapped context manager – if
    there is one.
    """

    def __init__(self, obj=None):
        self.obj = obj

    def __enter__(self):
        if self.obj is not None:
            return self.obj.__enter__()
        return None

    def __exit__(self, *tb):
        if self.obj is not None:
            return self.obj.__exit__(*tb)
        return None

    async def __aenter__(self):
        if self.obj is not None:
            return await self.obj.__aenter__()
        return None

    async def __aexit__(self, *tb):
        if self.obj is not None:
            return await self.obj.__aexit__(*tb)
        return None


def import_(name, off=0):
    """
    Import a module and access an object in it.

    `import_("a.b.c.d.e", 2)` imports "a.b.c" and returns the e attribute
    of object d from it.
    """
    n = name.split(".")
    mn = ".".join(n[: -off if off else 99])
    try:
        res = __import__(mn)
        for nn in n[1:]:
            res = getattr(res, nn)
    except Exception:
        sys.modules.pop(mn, None)
        raise
    return res


def load_from_cfg(*a, _cfg=None, _attr="server", _raise=False, **k):
    """
    A simple frontend to load a module, access a class/object from it,
    and call that with the config (and whchever other arguments you want to
    use).

    The module+object name is the "server" attribute (or @_attr).
    """
    if _cfg is None:
        cfg = k["cfg"]
    try:
        name = cfg[_attr]
    except KeyError:
        if _raise:
            raise
        return None
    if isinstance(name, (list, tuple)):
        name, off = name
    else:
        off = 1
    m = import_(name, off=off)
    return m(*a, **k)


def singleton(cls):
    """Basic singleton decorator"""
    return cls()


class TimeOnlyFormatter(logging.Formatter):
    """A log formatter that doesn't show dates"""

    default_time_format = "%H:%M:%S"
    default_msec_format = "%s.%03d"


def count(it):
    """counts the length of an iterator"""
    n = 0
    for _ in it:
        n += 1
    return n


async def acount(it):
    """counts the length of an async iterator"""
    n = 0
    async for _ in it:
        n += 1
    return n


class Cache:
    """
    A quick-and-dirty cache that keeps the last N entries of anything
    in memory so that ref and WeakValueDictionary don't lose them.

    Entries get refreshed when they're in the last third of the cache; as
    they're not removed, the actual cache size might only be 2/3rd of SIZE.
    """

    def __init__(self, size):
        self._size = size
        self._head = 0
        self._tail = 0
        self._attr = "_cache__pos"
        self._q = deque()

    def keep(self, entry):
        """Store an entry in the cache"""
        if getattr(entry, self._attr, -1) > self._tail + self._size / 3:
            return
        self._head += 1
        setattr(entry, self._attr, self._head)
        self._q.append(entry)
        self._flush()

    def _flush(self):
        while self._head - self._tail > self._size:
            self._q.popleft()
            self._tail += 1

    def resize(self, size):
        """Change the size of this cache."""
        self._size = size
        self._flush()

    def clear(self):
        """Clear the cache"""
        while self._head > self._tail:
            self._q.popleft()
            self._tail += 1


def digits(n, digits=6):  # pylint: disable=redefined-outer-name
    """
    Returns ``n`` rounded to ``digits`` significant digits. Default: 6.
    Ensures that the number doesn't carry nonsense precision or
    floating-point artefacts.

    >>> digits(123456789, 4)
    123400000
    >>> digits(math.pi, 4)
    3.142

    ``digits`` may be a fraction, in order to move the cut-off point to
    somewhere other than between 9.999 and 10.00.
    """
    return round(n, int(digits - 1 - log10(abs(n))))


def num2byte(num: int, length=None):
    """
    convert an unsigned integer to a bytestring
    """
    if length is None:
        length = (num.bit_length() + 7) // 8
    return num.to_bytes(length=length, byteorder="big")


def byte2num(data: bytes):
    """
    convert a bytestring to an unsigned integer
    """
    return int.from_bytes(data, byteorder="big")


def split_arg(p, kw):
    """
    Split argument 'p' and add to dict 'kw'.

    'p' may be of the form x=y (y is added, possibly as an integer),
    x? (call getpass(x? )), x?=y (call getpass(y: )).
    """
    try:
        k, v = p.split("=", 1)
    except ValueError:
        if p[-1] == "?":
            k = p[:-1]
            v = getpass(k + "? ")
        else:
            raise
    else:
        if k[-1] == "?":
            k = k[:-1]
            v = getpass(v + ": ")
        with suppress(ValueError):
            v = int(v)
    kw[k] = v


_alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"


def num2id(number, alphabet=_alphabet):
    """
    Encode a number / object ID in base36 (default).

    This code doesn't care about num2id(739172) or similar.

    To avoid these issues, pass an alphabet without vowels
    as the second parameter. `moat.util.random` contains some
    you might consider useful.
    """
    if not isinstance(number, int):
        if isinstance(number, (float, complex, str, bytes, bytearray)):
            raise TypeError("number must be an object or integer")
        number = id(number)
    is_negative = number < 0
    number = abs(number)
    res = []

    while number:
        number, i = divmod(number, len(alphabet))
        res.append(alphabet[i])
    if is_negative:
        res.append("-")
    elif not res:
        return alphabet[0]
    res.reverse()

    return "".join(res)



"""
Basic tool support
"""

from __future__ import annotations

import anyio
import logging  # pylint: disable=wrong-import-position
import io
import sys
from datetime import datetime
from functools import partial
from time import time
from collections.abc import Sequence

import asyncclick as click

from .main import load_subgroup
from .path import P, Path, path_eval, PathLongener,PathShortener
from .times import humandelta, time_until
from .yaml import yprint

log = logging.getLogger()


@load_subgroup(prefix="moat.util")
async def cli():
    """Various utilities"""
    # pylint:disable=unnecessary-pass


@cli.command(name="to")
@click.option("--sleep", "-s", is_flag=True, help="Sleep until that date/time")
@click.option("--human", "-h", is_flag=True, help="Print in numan-readable terms")
@click.option("--now", "-n", is_flag=True, help="Don't advance on match")
@click.option("--inv", "-i", is_flag=True, help="Time until no match")
@click.option("--back", "-b", is_flag=True, help="Time since the last (non)-match")
@click.option("--Human", "-H", "segments", type=int, default=None, help="Parts to print in 'human' mode")
@click.argument("args", nargs=-1)
async def to_(args, sleep, human, now, inv, back, segments):
    """\
Calculate the time until the start of the next given partial time
specification.

For instance, "moat util to 9 h": shows in how many seconds it's 9 o'clock
(possibly on the next day). Arbitrarily many units can be combined.

Negative numbers count from the end, i.e. "-2 hr" == 10 pm. Don't
forget to use the "--" option-to-argument separator if the time
specification starts with a negative number.

Days are numbered 1…7, Monday…Sunday. "3 dy" is synonymous to "wed",
while "3 wed" means "the third wednesday in a month".

"--human" prints a human-understandable version of the given
time. "--sleep" then delays until the specified moment arrives. If none
of these options is given, the number of seconds is printed.

By default, if the given time spec matches the current time, the
duration to the *next* moment the spec matches is calculated. Use
"--now" to print 0 / "now" / not sleep instead.

"--inv" inverts the time specification, i.e. "9 h" prints the time
until the next moment it is not / again no longer 9:** o'clock,
depending on whether "--now" is used / not used.

"--back" calculates the time *since the end* of the last match /
non-match instead. (If you want the start, use "--inv" and add a
second.)

\b
Known units:
s, sec  Second (0…59)
m, min  Minute (0…59)
h, hr   Hour (0…23)
d, dy   Day-of-week (1…7)
mon…sun Day-in-month (1…5)
w, wk   Week-of-year (0…53)
m, mo   Month (1…12)
y, yr   Year (2023–)
"""
    if not args:
        raise click.UsageError("Up to when please?")
    if back and sleep:
        raise click.UsageError("We don't have time machines.")

    t = datetime.now().astimezone()
    if not now:
        t = time_until(args, t, invert=not inv, back=back)
    t = time_until(args, t, invert=inv, back=back)

    t = t.timestamp()
    tt = int(t - time() + 0.9)
    if back:
        tt = -tt
    if human or segments:
        print(humandelta(tt, segments=segments or 3))
    if sleep:
        await anyio.sleep(tt)
    elif not human:
        print(tt)


@cli.command
@click.option("-d", "--dec", "--decoder", type=str, help="Source format", default="json")
@click.option("-e", "--enc", "--encoder", type=str, help="Destination format", default="yaml")
@click.option(
    "-i",
    "--in",
    "--input",
    "pathi",
    type=click.File("r"),
    help="Source file",
    default=sys.stdin,
)
@click.option(
    "-o",
    "--out",
    "--output",
    "patho",
    type=click.File("w"),
    help="Destination file",
    default=sys.stdout,
)
@click.option("-s", "--stream", is_flag=True, help="Multiple messages")
@click.option("-L", "--long", is_flag=True, help="Fix input paths")
@click.option("-S", "--short", is_flag=True, help="Compress output paths")
@click.option("-E", "--eval", "eval_", is_flag=True, help="Input line is a Python expr.")
@click.option("-D", "--dump", "dump_", is_flag=True, help="Output line is a Python repr.")
def convert(enc, dec, pathi, patho, stream, long, short, eval_, dump_):
    """File conversion utility.

    Supported file formats: json yaml cbor msgpack python std-cbor std-msgpack
    """

    if (long or short) and not stream:
        raise click.UsageError("Path mods only make sense when streaming")
    if long and short:
        raise click.UsageError("Both Path mods together are a no-op")
    if long:
        long=PathLongener()
    if short:
        long=PathShortener()

    class IT:
        def __init__(self, codec):
            self.codec = codec

        def __call__(self, buf):
            if eval_:
                buf = eval(buf)
            self.codec.feed(buf)
            return iter(self.codec)

    class EV:
        def __init__(self, codec):
            self.codec = codec

        def __call__(self, buf):
            buf = eval(buf)
            return self.codec(buf)

    def get_codec(n):
        if n == "python":
            from pprint import pformat

            return eval, pformat, False, False

        if n == "json":
            import simplejson as json

            if stream:

                def jdump(d):
                    return json.dumps(d, separators=(",", ":")) + "\n"

            else:

                def jdump(d):
                    if dump_:
                        res = json.dumps(d, separators=(",", ":"), indent="")
                        return repr(res)
                    return json.dumps(d, indent="  ") + "\n"

            return json.loads, jdump, False, False

        if n == "yaml":
            import ruyaml as yaml

            y = yaml.YAML(typ="safe")
            y.default_flow_style = True, False
            from moat.util import yload

            def ypr(d, s):
                if dump_:
                    buf = io.StringIO()
                    yprint(d, buf)
                    print(repr(buf.getvalue()), file=s)
                else:
                    yprint(d, s)
                    s.write("---\n")

            return (
                partial(yload, multi=True) if stream else yload,
                ypr if stream else yprint,
                False,
                True,
            )

        if n == "cbor":
            from moat.lib.codec.cbor import Codec as CBOR

            c = CBOR()
            d = CBOR()
            return IT(c) if stream else EV(c.decode) if eval_ else c.decode, d.encode, True, False

        if n == "std-cbor":
            from moat.util.cbor import StdCBOR

            c = StdCBOR()
            d = StdCBOR()
            return IT(c) if stream else EV(c.decode) if eval_ else c.decode, d.encode, True, False

        if n == "msgpack":
            from moat.lib.codec.msgpack import Codec as Msgpack

            c = Msgpack()
            d = Msgpack()
            return IT(c) if stream else EV(c.decode) if eval_ else c.decode, d.encode, True, False

        if n == "std-msgpack":
            from moat.util.msgpack import StdMsgpack

            c = StdMsgpack()
            d = StdMsgpack()
            return IT(c) if stream else EV(c.decode) if eval_ else c.decode, d.encode, True, False

        raise ValueError("unsupported codec")

    dec, _x, bd, csd = get_codec(dec)
    _y, enc, be, cse = get_codec(enc)
    if bd:
        pathi = pathi.buffer
    if be:
        if dump_:
            bt = io.BytesIO
        else:
            patho = patho.buffer
    else:
        if dump_:
            bt = io.StringIO

    if stream:
        if csd:

            def in_d():
                return [pathi.read()]

        else:

            def in_d():
                while data := pathi.read(4096):
                    yield data

        for d in in_d():
            for data in dec(d):
                if long and isinstance(data,Sequence):
                    d,p,*x = data
                    p = long.long(d,p)
                    data = [p,*x]
                if short and isinstance(data,Sequence):
                    p,*x = data
                    d,p = short.short(p)
                    data = [d,p,*x]

                if cse:
                    if dump_:
                        buf = bt()
                        enc(data, buf)
                        patho.write(repr(buf.getvalue()) + "\n")
                    else:
                        enc(data, patho)
                else:
                    dat = enc(data)
                    if dump_:
                        dat = repr(dat)
                    patho.write(dat)
    else:
        if csd:
            data = dec(pathi)
        else:
            data = pathi.read()
            data = dec(data)
        if cse:
            if dump_:
                buf = bt()
                enc(data, buf)
                patho.write(repr(buf.getvalue()) + "\n")
            else:
                enc(data, patho)
        else:
            data = enc(data)
            if dump_:
                data = repr(data)
            patho.write(data)
    pathi.close()
    patho.close()


@cli.command("path", help=Path.__doc__, no_args_is_help=True)
@click.option(
    "-e",
    "--encode",
    is_flag=True,
    help="evaluate a Python expr and encode to a pathstr",
)
@click.option("-d", "--decode", is_flag=True, help="decode a path to a list")
@click.argument("path", nargs=-1)
async def path_(encode, decode, path):
    """Explain/test MoaT paths"""
    if not encode and not decode:
        raise click.UsageError("Need -e or -d option.")
    elif not decode:
        try:
            path = path_eval(" ".join(path))
        except Exception as exc:  # pylint:disable=broad-exception-caught
            print(repr(exc), file=sys.stderr)
        if not isinstance(path, (list, tuple)):
            path = path[0]
        print(Path(*path))
    elif encode:
        raise click.UsageError("encode and decode at the same time??")
    else:
        for p in path:
            print(repr(list(P(p))))


@cli.command("cfg", help="Retrieve+show a config value", no_args_is_help=True)
@click.pass_obj
@click.option("-y", "--yaml", is_flag=True, help="print as YAML")
@click.option("-e", "--empty", is_flag=True, help="empty string if the key doesn't exist")
@click.argument("path", nargs=-1, type=P)
async def cfg_(obj, path, yaml, empty):
    """Emit the current configuration as a YAML file.

    You can limit the output by path elements.
    E.g., "cfg kv.connect.host" will print "localhost".

    Single values are printed with a trailing line feed.

    Dump the whole config with "moat util cfg :".
    """
    from .exc import ungroup

    delim = False
    for p in path:
        if delim and yaml:
            print("---", file=obj.stdout)
        with ungroup():
            try:
                v = obj.cfg._get(p)  # noqa:SLF001
            except KeyError:
                if not empty:
                    print("Unknown:", p, file=sys.stderr)
                    sys.exit(1)
            else:
                if yaml:
                    yprint(v, obj.stdout)
                else:
                    print(v, file=obj.stdout)
        delim = True



"""
Compatibility wrappers that allows MoaT code to run on CPython/anyio as
well as MicroPython/uasyncio.

Well, for the most part.
"""

from __future__ import annotations

import anyio as _anyio
import logging
import os
import sys
import time as _time
import traceback as _traceback
from concurrent.futures import CancelledError
from contextlib import suppress, AsyncExitStack
from inspect import currentframe, iscoroutinefunction, iscoroutine
from codecs import utf_8_decode

from .queue import Queue as _Queue
from .queue import QueueEmpty, QueueFull
from moat.util.merge import merge

logger = logging.getLogger(__name__)

__all__ = [
    "is_async",
    "doc",
    "log",
    "at",
    "const",
    "CancelScope",
    "Queue",
    "print_exc",
    "ticks_ms",
    "sleep_ms",
    "wait_for",
    "wait_for_ms",
    "every_ms",
    "every",
    "idle",
    "ticks_add",
    "ticks_diff",
    "run",
    "byte2utf8",
    "TaskGroup",
    "run_server",
    "shield",
    "Event",
    "Lock",
    "L",
    "WouldBlock",
]


Pin_IN = 0
Pin_OUT = 1

Event = _anyio.Event
Lock = _anyio.Lock
WouldBlock = _anyio.WouldBlock
sleep = _anyio.sleep
EndOfStream = _anyio.EndOfStream
BrokenResourceError = _anyio.BrokenResourceError
ClosedResourceError = _anyio.ClosedResourceError
TimeoutError = TimeoutError  # noqa:PLW0127,A001
ExceptionGroup = ExceptionGroup  # noqa: PLW0127
BaseExceptionGroup = BaseExceptionGroup  # noqa: PLW0127


def const(_x):
    "µPython compatibility"
    return _x


L = True

Pin_IN = 0
Pin_OUT = 1


def byte2utf8(buf: bytes | bytearray | memoryview) -> str:
    res, n = utf_8_decode(buf)
    if n != len(buf):
        raise ValueError("incomplete utf8")
    return res


class CancelScope:
    """
    An async-await-able CancelScope wrapper
    """

    def __init__(self):
        self.sc = _anyio.CancelScope()

    async def __aenter__(self):
        self.sc.__enter__()
        return self

    async def __aexit__(self, *tb):
        return self.sc.__exit__(*tb)

    def cancel(self):
        "Cancel the scope"
        self.sc.cancel()

    @property
    def cancelled(self):
        "Was 'cancel' called on this scope?"
        return self.sc.cancel_called()


def log(s, *x, err=None, nback=1):
    "Basic logger.debug/error call (depends on @err)"
    caller = currentframe()
    for _ in range(nback):
        if caller.f_back is None:
            break
        caller = caller.f_back
    log_ = logging.getLogger(caller.f_globals["__name__"])
    (log_.debug if err is None else log_.error)(s, *x, exc_info=err, stacklevel=1 + nback)
    if err and int(os.getenv("LOG_BRK", "0")):
        breakpoint()  # noqa:T100 pylint:disable=forgotten-debug-statement

def at(*a,**kw):
    log_.debug("%r %r",a,kw)

def print_exc(exc, file=None):
    "print a stack trace to stderr"
    _traceback.print_exception(type(exc), exc, exc.__traceback__, file=file)


def ticks_ms():
    "return a monotonic timer, in milliseconds"
    return _time.monotonic_ns() // 1000000


async def sleep_ms(ms):
    "sleep for @ms milliseconds"
    await sleep(ms / 1000)


async def wait_for(timeout, p, *a, **k):
    "timeout if the call to p(*a,**k) takes longer than @timeout seconds"
    with _anyio.fail_after(timeout):
        return await p(*a, **k)


async def wait_for_ms(timeout, p, *a, **k):
    "timeout if the call to p(*a,**k) takes longer than @timeout milliseconds"
    with _anyio.fail_after(timeout / 1000):
        return await p(*a, **k)


async def every_ms(t, p, *a, **k):
    "every t milliseconds, call p(*a,**k)"
    tt = ticks_add(ticks_ms(), t)
    while True:
        try:
            yield await p(*a, **k)
        except StopAsyncIteration:
            return
        tn = ticks_ms()
        td = ticks_diff(tt, tn)
        if td > 0:
            await sleep_ms(td)
            tt += t
        else:
            # owch, delay too long
            tt = ticks_add(tn, t)


def every(t, p, *a, **k):
    "every t seconds, call p(*a,**k)"
    return every_ms(t * 1000, p, *a, **k)


async def idle():
    "sleep forever"
    while True:
        await _anyio.sleep(60 * 60 * 12)  # half a day


def ticks_add(a, b):
    "returns a+b"
    return a + b


def ticks_diff(a, b):
    "returns a-b"
    return a - b


def run(p, *a, **k):
    "wrapper for anyio.run"
    return _anyio.run(p, a, k)


_tg = None
_tgt = None


def TaskGroup():
    "A TaskGroup subclass (generator) that supports `spawn` and `cancel`"

    global _tg, _tgt  # noqa:PLW0603 pylint:disable=global-statement
    if "pytest" in sys.modules or _tgt is None:
        tgt = type(_anyio.create_task_group())
    else:
        tgt = _tgt
    if tgt is not _tgt:
        _tgt = tgt

        class TaskGroup_(_tgt):
            """An augmented taskgroup"""

            async def spawn(self, p, *a, _name=None, **k):
                """
                Like start(), but returns something you can cancel
                """
                # logger.info("Launch %s %s %s %s",_name, p,a,k)

                async def catch(p, a, k, *, task_status):
                    with _anyio.CancelScope() as s:
                        task_status.started(s)
                        with suppress(CancelledError):  # error from concurrent.futures
                            await p(*a, **k)

                return await super().start(catch, p, a, k, name=_name)

            def cancel(self):
                "cancel all tasks in this taskgroup"
                self.cancel_scope.cancel()

        _tg = TaskGroup_
    return _tg()


async def run_server(cb, host, port, backlog=5, taskgroup=None, reuse_port=True, evt=None):
    """Listen to and serve a TCP stream.

    This mirrors [u]asyncio, except that the callback gets the socket once.
    """
    listener = await _anyio.create_tcp_listener(
        local_host=host,
        local_port=port,
        backlog=backlog,
        reuse_port=reuse_port,
    )
    async with listener:
        if evt is not None:
            evt.set()
        await listener.serve(cb, task_group=taskgroup)


def shield():
    """A wrapper shielding the contents from external cancellation.

    Equivalent to ``CancelScope(shield=True)``.
    """
    return _anyio.CancelScope(shield=True)


class Queue(_Queue):
    """
    compatibility mode: raise `EOFError` and `QueueEmpty`/`QueueFull`
    instead of `anyio.EndOfStream` and `anyio.WouldBlock`
    """

    async def get(self):  # noqa:D102
        try:
            return await super().get()
        except _anyio.EndOfStream:
            raise EOFError from None

    def get_nowait(self):  # noqa:D102
        try:
            return super().get_nowait()
        except _anyio.EndOfStream:
            raise EOFError from None
        except _anyio.WouldBlock:
            raise QueueEmpty from None

    def put_nowait(self, x):  # noqa:D102
        try:
            super().put_nowait(x)
        except _anyio.WouldBlock:
            raise QueueFull from None


def _doc(_c=None, **kw):
    """
    Attach structured documentation to a function.

    This is used for command handlers because we want to (a) not add a heap
    of obscure typing to the Message parameter that'd need to be serialozed
    somehow, (b) support introspectable doc for MicroPython which has no typing
    infrastructure.

    Keywords:
    _c: copy+extend Upstream
    _d: very short Documentation, free text
    _r: Return value
    _i: Input stream type
    _o: Output stream type
    _a: Any keyword params
    _m: first optional field
    _NUM: positional arg
    _99: positional arg trailer
    NAME: keyword arg

    All values are of the form `type`, `type:short documentation`, a dict
    (as above, except no NUMs), or a list (positional values, like typing
    w/ tuples). A trailing question mark on the type indicates "or None".
    (This doesn't necessarily mean the same as not sending the value at all.)

    Typically the return value is a single data item; if not you can set
    '_r' to a dict with the above components.

    The list of possibly-missing fields may include "i" and "o".
    If any positional arguments are optional, only the first should be given.
    Keyword args are always optional.

    The result type "parts" describes a two-element tuple: a dict/list with
    complex items removed plus a list of sub-keys to retrieve for restoring
    the data. This is used to limit the max block size.
    """
    if _c is not None:
        merge(kw, _c._moat__doc, replace=False)

    def mod(fn):
        fn._moat__doc = kw
        return fn

    return mod


# async context stack


def ACM(obj):
    """A bare-bones async context manager / async exit stack.

    Usage::

        class Foo():
            async def __aenter__(self):
                AC = ACM(obj)
                try:
                    ctx1 = await AC(obj1)
                    ctx2 = await AC_use(self, obj2)  # same thing
                    ...
                    return self_or_whatever

                except BaseException:
                    await AC_exit(self, *exc)
                    raise

            async def __aexit__(self, *exc):
                return await AC_exit(self, *exc)

    Calls to `ACM` and `AC_exit` can be nested, even on the same object.
    They **must** balance, hence the above error handling dance.
    """
    # pylint:disable=protected-access
    if not hasattr(obj, "_AC_"):
        obj._AC_ = []

    cm = AsyncExitStack()
    obj._AC_.append(cm)

    # AsyncExitStack.__aenter__ is a no-op. We don't depend on that but at
    # least it shouldn't yield
    # log("AC_Enter",nback=2)
    try:
        # pylint:disable=no-member,unnecessary-dunder-call
        cr = cm.__aenter__()
        cr.send(None)
    except StopIteration as s:
        cm = s.value
    else:
        raise RuntimeError("AExS ??")

    def _ACc(ctx):
        return AC_use(obj, ctx)

    return _ACc


async def AC_use(obj, ctx):
    """
    Add an async context to this object's AsyncExitStack.

    If the object is a context manager (async or sync), this opens the
    context and return its value.

    Otherwise it's a callable and will run on exit.
    """
    acm = obj._AC_[-1]
    if hasattr(ctx, "__aenter__"):
        return await acm.enter_async_context(ctx)
    elif hasattr(ctx, "__enter__"):
        return acm.enter_context(ctx)
    elif iscoroutinefunction(ctx):
        acm.push_async_callback(ctx)
    elif iscoroutine(ctx):
        raise ValueError(ctx)
    else:
        acm.callback(ctx)
    return None


async def AC_exit(obj, *exc):
    """End the latest AsyncExitStack opened by `ACM`."""
    if not exc:
        exc = (None, None, None)
    return await obj._AC_.pop().__aexit__(*exc)


def is_async(obj):
    if hasattr(obj, "__await__"):
        return True
    return False
